import numpy as np
import pytest

import sdr


def test_amplitude():
    x = np.exp(1j * 2 * np.pi * 5 * np.arange(100) / 100)
    amplitude = 5  # dB
    y = sdr.iq_imbalance(x, amplitude)

    before = 10 * np.log10(sdr.average_power(x.real) / sdr.average_power(x.imag))
    after = 10 * np.log10(sdr.average_power(y.real) / sdr.average_power(y.imag))
    assert amplitude == pytest.approx(after - before)


# def test_phase():
#     # TODO: How to test phase imbalance?
#     return



import numpy as np

import sdr


def test_code_to_sequence():
    code = np.array([0, 1, 1, 0])
    sequence = sdr._sequence._conversion.code_to_sequence(code)
    sequence_truth = np.array([1, -1, -1, 1])
    assert np.array_equal(sequence, sequence_truth)


def test_sequence_to_code():
    sequence = np.array([1, -1, -1, 1])
    code = sdr._sequence._conversion.sequence_to_code(sequence)
    code_truth = np.array([0, 1, 1, 0])
    assert np.array_equal(code, code_truth)



import pytest

import sdr


def test_accumulate():
    ber = sdr.ErrorRate()
    x = [1, 1, 1, 1, 1]  # Reference bit vector

    x_hat = [1, 0, 1, 1, 1]  # Received bit vector
    assert ber.add(10, x, x_hat) == (1, 5, 0.2)
    assert ber.errors(10) == 1
    assert ber.counts(10) == 5
    assert ber.error_rate(10) == 0.2

    x_hat = [1, 0, 1, 0, 1]  # Received bit vector
    assert ber.add(10, x, x_hat) == (2, 5, 0.4)
    assert ber.errors(10) == 3
    assert ber.counts(10) == 10
    assert ber.error_rate(10) == 0.3


def test_no_snr():
    ber = sdr.ErrorRate()

    assert ber.errors(10) == 0
    assert ber.counts(10) == 0
    with pytest.raises(ZeroDivisionError):
        assert ber.error_rate(10)



import numpy as np
import pytest

import sdr


def test_power():
    rng = np.random.default_rng()
    A = rng.uniform(2, 10)  # Signal amplitude
    N = 50
    x = A * np.exp(1j * np.linspace(0, 2 * np.pi, N, endpoint=False))
    assert np.allclose(sdr.power(x), A**2)


def test_peak_power():
    rng = np.random.default_rng()
    A = rng.uniform(2, 10)  # Signal amplitude
    N = 50
    x = A * np.sin(np.linspace(0, 2 * np.pi, N, endpoint=False))
    assert sdr.peak_power(x) == pytest.approx(A**2, rel=1e-2)


def test_average_power():
    rng = np.random.default_rng()
    A = rng.uniform(2, 10)  # Signal amplitude
    N = 50
    x = A * np.sin(np.linspace(0, 2 * np.pi, N, endpoint=False))
    assert sdr.average_power(x) == pytest.approx(A**2 / 2, rel=1e-2)



"""
A module containing various power measurement functions.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from .._conversion import db as to_db
from .._helper import convert_output, export, verify_arraylike, verify_bool


@export
def power(
    x: npt.ArrayLike,
    db: bool = False,
) -> float:
    r"""
    Measures the instantaneous power of a time-domain signal $x[n]$.

    $$P = \left| x[n] \right|^2$$

    Arguments:
        x: The time-domain signal $x[n]$ to measure.
        db: Indicates whether to return the result in decibels (dB).

    Returns:
        The instantaneous power. If `db=False`, $P$ is returned.
        If `db=True`, $10 \log_{10} P$ is returned.

    Group:
        measurement-power
    """
    x = verify_arraylike(x, complex=True)
    verify_bool(db)

    P = np.abs(x) ** 2
    if db:
        P = to_db(P, type="power")

    return convert_output(P)


@export
def peak_power(
    x: npt.ArrayLike,
    axis: int | tuple[int, ...] | None = None,
    db: bool = False,
) -> float:
    r"""
    Measures the peak power of a time-domain signal $x[n]$.

    $$P_{\text{peak}} = \max \left| x[n] \right|^2$$

    Arguments:
        x: The time-domain signal $x[n]$ to measure.
        axis: Axis or axes along which to compute the peak power. The default is `None`, which computes the peak power
            of the entire array.
        db: Indicates whether to return the result in decibels (dB).

    Returns:
        The peak power. If `db=False`, $P_{\text{peak}}$ is returned.
        If `db=True`, $10 \log_{10} P_{\text{peak}}$ is returned.

    Group:
        measurement-power
    """
    x = verify_arraylike(x, complex=True)
    verify_bool(db)

    P_peak = np.max(np.abs(x) ** 2, axis=axis)
    if db:
        P_peak = to_db(P_peak, type="power")

    return convert_output(P_peak)


@export
def average_power(
    x: npt.ArrayLike,
    axis: int | tuple[int, ...] | None = None,
    db: bool = False,
) -> float:
    r"""
    Measures the average power of a time-domain signal $x[n]$.

    $$P_{\text{avg}} = \frac{E}{N} = \frac{1}{N} \sum_{n=0}^{N-1} \left| x[n] \right|^2$$

    Arguments:
        x: The time-domain signal $x[n]$ to measure.
        axis: Axis or axes along which to compute the average power. The default is `None`, which computes the average
            power of the entire array.
        db: Indicates whether to return the result in decibels (dB).

    Returns:
        The average power. If `db=False`, $P_{\text{avg}}$ is returned.
        If `db=True`, $10 \log_{10} P_{\text{avg}}$ is returned.

    Group:
        measurement-power
    """
    x = verify_arraylike(x, complex=True)

    P_avg = np.mean(np.abs(x) ** 2, axis=axis)
    if db:
        P_avg = to_db(P_avg, type="power")

    return convert_output(P_avg)


@export
def papr(
    x: npt.ArrayLike,
    axis: int | tuple[int, ...] | None = None,
) -> float:
    r"""
    Measures the peak-to-average power ratio (PAPR) of a time-domain signal $x[n]$.

    $$\text{PAPR} = 10 \log_{10} \frac{P_{\text{peak}}}{P_{\text{avg}}}$$

    Arguments:
        x: The time-domain signal $x[n]$ to measure.
        axis: Axis or axes along which to compute the PAPR. The default is `None`, which computes the PAPR of the
            entire array.

    Returns:
        The PAPR of $x[n]$ in dB.

    See Also:
        sdr.peak_power, sdr.average_power

    References:
        - https://en.wikipedia.org/wiki/Crest_factor

    Group:
        measurement-power
    """
    return to_db(peak_power(x, axis=axis) / average_power(x, axis=axis))



import numpy as np

import sdr


def test_da_error():
    A_rx, A_ref = 5, 3
    ped = sdr.MLPED(A_rx, A_ref)
    modem = sdr.PSK(4)
    error, da_error = ped.data_aided_error(modem)
    da_error_truth = A_rx * A_ref * np.sin(error)
    assert np.allclose(da_error, da_error_truth)



import numpy as np
import pytest

import sdr


def test_scalar():
    assert sdr.max_frequency_offset(3, 1e-3) == pytest.approx(442.2433896262681)


def test_cgl_vector():
    f = sdr.max_frequency_offset([1, 2, 3, 4, 5], 1e3)
    f_truth = np.array([0.0002615, 0.00036547, 0.00044224, 0.00050444, 0.00055699])
    assert isinstance(f, np.ndarray)
    assert np.allclose(f, f_truth)


def test_freq_vector():
    f = sdr.max_frequency_offset(3, [1e-3, 2e-3, 3e-3, 4e-3, 5e-3])
    f_truth = np.array([442.24338963, 221.12169481, 147.41446321, 110.56084741, 88.44867793])
    assert isinstance(f, np.ndarray)
    assert np.allclose(f, f_truth)



"""
A module containing frequency-domain plotting functions.
"""

from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import scipy.fft
import scipy.signal
from typing_extensions import Literal

from .._helper import export, verify_arraylike, verify_bool, verify_isinstance, verify_literal, verify_scalar
from ._helper import integer_x_axis, standard_plot, verify_sample_rate
from ._rc_params import RC_PARAMS


@export
def dft(
    x: npt.ArrayLike,
    sample_rate: float | None = None,
    window: str | float | tuple | None = None,
    size: int | None = None,
    oversample: int | None = None,
    fast: bool = False,
    centered: bool = True,
    ax: plt.Axes | None = None,
    type: Literal["plot", "stem"] = "plot",
    x_axis: Literal["freq", "bin"] = "freq",
    y_axis: Literal["complex", "mag", "mag^2", "db"] = "mag",
    diff: Literal["color", "line"] = "color",
    **kwargs,
):
    r"""
    Plots the discrete Fourier transform (DFT) of the time-domain signal $x[n]$.

    Arguments:
        x: The time-domain signal $x[n]$.
        sample_rate: The sample rate $f_s$ of the signal in samples/s. If `None`, the x-axis will
            be labeled as "Normalized frequency".
        window: The SciPy window definition. See :func:`scipy.signal.windows.get_window` for details.
            If `None`, no window is applied.
        size: The number of points to use for the DFT. If `None`, the length of the signal is used.
        oversample: The factor to oversample the DFT. If `None`, the DFT is not oversampled. This is only considered
            if `size` is `None`.
        fast: Indicates whether to use the fast Fourier transform (FFT) algorithm. If `True`, the DFT size is set
            to the next power of 2.
        centered: Indicates whether to center the DFT about 0.
        ax: The axis to plot on. If `None`, the current axis is used.
        type: The type of plot to use.
        x_axis: The x-axis scaling.
        y_axis: The y-axis scaling.
        diff: Indicates how to differentiate the real and imaginary parts of a complex signal. If `"color"`, the
            real and imaginary parts will have different colors based on the current Matplotlib color cycle.
            If `"line"`, the real part will have a solid line and the imaginary part will have a dashed line,
            and both lines will share the same color.
        kwargs: Additional keyword arguments to pass to the plotting function.

    See Also:
        sdr.plot.dtft

    Notes:
        The discrete Fourier transform (DFT) is defined as

        $$X[k] = \sum_{n=0}^{N-1} x[n] e^{-j 2 \pi k n / N},$$

        where $x[n]$ is the time-domain signal, $X[k]$ is the DFT, $k$ is the DFT frequency bin, and $N$ is the number
        of samples in the DFT. The frequency corresponding to the k-th bin is $\frac{k}{N} f_s$.

        The DFT is a sampled version of the discrete-time Fourier transform (DTFT).

    Examples:
        Create a tone whose frequency will straddle two DFT bins.

        .. ipython:: python

            n = 10  # samples
            f = 1.5 / n  # cycles/sample
            x = np.exp(1j * 2 * np.pi * f * np.arange(n))

            @savefig sdr_plot_dft_1.png
            plt.figure(); \
            sdr.plot.dft(x, x_axis="bin", type="stem");

        Plot the DFT against normalized frequency. Compare the DTFT, an oversampled DFT, and a critically sampled DFT.
        Notice that the critically sampled DFT has scalloping loss (difference between the true peak and the DFT peak)
        when the signal frequency does not align with a bin. Furthermore, the sidelobes are non-zero and large.
        This is known as spectral leakage -- the spreading of a tone's power from a single bins to all bins.

        .. ipython:: python

            @savefig sdr_plot_dft_2.png
            plt.figure(); \
            sdr.plot.dtft(x, color="k", label="DTFT"); \
            sdr.plot.dft(x, oversample=4, type="stem", label="4x oversampled DFT"); \
            sdr.plot.dft(x, type="stem", label="DFT");

        If a window is applied to the signal before the DFT is computed, the main lobe is widened and the sidelobes are
        reduced. Given the wider main lobe, the scalloping loss of the critically sampled DFT is also reduced. These
        benefits come at the cost of reduced frequency resolution.

        .. ipython:: python

            @savefig sdr_plot_dft_3.png
            plt.figure(); \
            sdr.plot.dtft(x, window="hamming", color="k", label="DTFT"); \
            sdr.plot.dft(x, oversample=4, window="hamming", type="stem", label="4x oversampled DFT"); \
            sdr.plot.dft(x, window="hamming", type="stem", label="DFT");

    Group:
        plot-frequency-domain
    """
    x = verify_arraylike(x, complex=True, ndim=1)
    sample_rate, sample_rate_provided = verify_sample_rate(sample_rate)
    # verify_isinstance(window, str, optional=True)
    verify_scalar(size, optional=True, int=True, positive=True)
    verify_scalar(oversample, optional=True, int=True, positive=True)
    verify_bool(fast)
    verify_bool(centered)
    verify_isinstance(ax, plt.Axes, optional=True)
    verify_literal(type, ["plot", "stem"])
    verify_literal(x_axis, ["freq", "bin"])
    verify_literal(y_axis, ["complex", "mag", "mag^2", "db"])
    verify_literal(diff, ["color", "line"])

    with plt.rc_context(RC_PARAMS):
        if ax is None:
            ax = plt.gca()

        if window is not None:
            x *= scipy.signal.windows.get_window(window, x.size)

        if size is None:
            if oversample is None:
                size = x.size
            else:
                size = x.size * oversample

        if fast:
            size = scipy.fft.next_fast_len(size)

        X = np.fft.fft(x, size)
        if x_axis == "freq":
            f = np.fft.fftfreq(size, 1 / sample_rate)
        elif x_axis == "bin":
            f = np.fft.fftfreq(size, 1 / size)

        if centered:
            X = np.fft.fftshift(X)
            f = np.fft.fftshift(f)

        standard_plot(f, X, ax=ax, y_axis=y_axis, diff=diff, type=type, **kwargs)

        if x_axis == "bin":
            integer_x_axis(ax)
            ax.set_xlabel("DFT bin index, $k$")
        elif sample_rate_provided:
            ax.set_xlabel("Frequency (Hz), $f$")
        else:
            ax.set_xlabel("Normalized frequency, $f/f_s$")

        if y_axis == "complex":
            ax.set_ylabel("Amplitude, $X[k]$")
        elif y_axis == "mag":
            ax.set_ylabel(r"Magnitude, $\left| X[k] \right|$")
        elif y_axis == "mag^2":
            ax.set_ylabel(r"Power, $\left| X[k] \right|^2$")
        elif y_axis == "db":
            ax.set_ylabel(r"Power (dB), $\left| X[k] \right|^2$")

        ax.set_title("Discrete Fourier transform (DFT)")


@export
def dtft(
    x: npt.ArrayLike,
    sample_rate: float | None = None,
    window: str | float | tuple | None = None,
    size: int = int(2**20),  # ~1 million points
    centered: bool = True,
    ax: plt.Axes | None = None,
    y_axis: Literal["complex", "mag", "mag^2", "db"] = "mag",
    diff: Literal["color", "line"] = "color",
    **kwargs,
):
    r"""
    Plots the discrete-time Fourier transform (DTFT) of the time-domain signal $x[n]$.

    Arguments:
        x: The time-domain signal $x[n]$.
        sample_rate: The sample rate $f_s$ of the signal in samples/s. If `None`, the x-axis will
            be labeled as "Normalized frequency".
        window: The SciPy window definition. See :func:`scipy.signal.windows.get_window` for details.
            If `None`, no window is applied.
        size: The number of points to use for the DTFT. The actual size used will be the nearest power of 2.
        centered: Indicates whether to center the DTFT about 0.
        ax: The axis to plot on. If `None`, the current axis is used.
        y_axis: The y-axis scaling.
        diff: Indicates how to differentiate the real and imaginary parts of a complex signal. If `"color"`, the
            real and imaginary parts will have different colors based on the current Matplotlib color cycle.
            If `"line"`, the real part will have a solid line and the imaginary part will have a dashed line,
            and both lines will share the same color.
        kwargs: Additional keyword arguments to pass to the plotting function.

    See Also:
        sdr.plot.dft

    Notes:
        The discrete Fourier transform (DTFT) is defined as

        $$X(f) = \sum_{n=-\infty}^{\infty} x[n] e^{-j 2 \pi f n / f_s},$$

        where $x[n]$ is the time-domain signal, $X(f)$ is the DTFT, and $f$ is the frequency.

    Examples:
        Create a DC tone that is 10 samples long. Plot its DTFT. Notice that the width of the main lobe is $2 / T$,
        with nulls at $\pm 1 / T$.

        .. ipython:: python

            n = 10  # samples
            f = 0 / n  # cycles/sample
            x = np.exp(1j * 2 * np.pi * f * np.arange(n))

            @savefig sdr_plot_dtft_1.png
            plt.figure(); \
            sdr.plot.dtft(x);

        Plot a critically sampled DFT and an oversampled DFT of the signal. Notice that the DFT is a sampled version of
        the DTFT. The oversampled DFT has more samples and thus more closely resembles the DTFT.

        .. ipython:: python

            @savefig sdr_plot_dtft_2.png
            plt.figure(); \
            sdr.plot.dtft(x, label="DTFT"); \
            sdr.plot.dft(x, oversample=4, type="stem", label="4x oversampled DFT"); \
            sdr.plot.dft(x, type="stem", label="DFT");

    Group:
        plot-frequency-domain
    """
    with plt.rc_context(RC_PARAMS):
        dft(
            x,
            sample_rate=sample_rate,
            window=window,
            size=size,
            fast=True,
            centered=centered,
            ax=ax,
            type="plot",
            x_axis="freq",
            y_axis=y_axis,
            diff=diff,
            **kwargs,
        )

        ax = plt.gca()

        if y_axis == "complex":
            ax.set_ylabel("Amplitude, $X(f)$")
        elif y_axis == "mag":
            ax.set_ylabel(r"Magnitude, $\left| X(f) \right|$")
        elif y_axis == "mag^2":
            ax.set_ylabel(r"Power, $\left| X(f) \right|^2$")
        elif y_axis == "db":
            ax.set_ylabel(r"Power (dB), $\left| X(f) \right|^2$")

        ax.set_title("Discrete-time Fourier transform (DTFT)")



"""
A module containing a Farrow arbitrary resampler.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt
import scipy.signal

from ._helper import convert_output, export, verify_arraylike, verify_bool, verify_scalar


@export
class FarrowResampler:
    r"""
    Implements a piecewise polynomial Farrow arbitrary resampler.

    References:
        - Michael Rice, *Digital Communications: A Discrete Time Approach*, Section 8.4.2.
        - `Qasim Chaudhari, Fractional Delay Filters Using the Farrow Structure.
          <https://wirelesspi.com/fractional-delay-filters-using-the-farrow-structure/>`_

    Examples:
        Create a sine wave with angular frequency $\omega = 2 \pi / 5.179$. Interpolate the signal by
        $r = \pi$ using Farrow piecewise polynomial Farrow resamplers.

        .. ipython:: python

            x = np.cos(2 * np.pi / 5.179 * np.arange(11))
            rate = np.pi

        Create a linear Farrow piecewise polynomial interpolator.

        .. ipython:: python

            farrow1 = sdr.FarrowResampler(1)
            y1 = farrow1(x, rate)

            @savefig sdr_FarrowResampler_1.png
            plt.figure(); \
            sdr.plot.time_domain(x, sample_rate=1, marker="o", label="Input"); \
            sdr.plot.time_domain(y1, sample_rate=rate, marker=".", label="Output"); \
            plt.title("Linear Farrow Resampler");

        Create a quadratic Farrow piecewise polynomial interpolator.

        .. ipython:: python

            farrow2 = sdr.FarrowResampler(2)
            y2 = farrow2(x, rate)

            @savefig sdr_FarrowResampler_2.png
            plt.figure(); \
            sdr.plot.time_domain(x, sample_rate=1, marker="o", label="Input"); \
            sdr.plot.time_domain(y2, sample_rate=rate, marker=".", label="Output"); \
            plt.title("Quadratic Farrow Resampler");

        Create a cubic Farrow piecewise polynomial interpolator.

        .. ipython:: python

            farrow3 = sdr.FarrowResampler(3)
            y3 = farrow3(x, rate)

            @savefig sdr_FarrowResampler_3.png
            plt.figure(); \
            sdr.plot.time_domain(x, sample_rate=1, marker="o", label="Input"); \
            sdr.plot.time_domain(y3, sample_rate=rate, marker=".", label="Output"); \
            plt.title("Cubic Farrow Resampler");

        Create a quartic Farrow piecewise polynomial interpolator.

        .. ipython:: python

            farrow4 = sdr.FarrowResampler(4)
            y4 = farrow4(x, rate)

            @savefig sdr_FarrowResampler_4.png
            plt.figure(); \
            sdr.plot.time_domain(x, sample_rate=1, marker="o", label="Input"); \
            sdr.plot.time_domain(y4, sample_rate=rate, marker=".", label="Output"); \
            plt.title("Quartic Farrow Resampler");

        Compare the outputs of the Farrow resamplers with varying polynomial order.

        .. ipython:: python

            @savefig sdr_FarrowResampler_5.png
            plt.figure(); \
            sdr.plot.time_domain(x, sample_rate=1, marker="o", label="Input"); \
            sdr.plot.time_domain(y1, sample_rate=rate, marker=".", label="Linear"); \
            sdr.plot.time_domain(y2, sample_rate=rate, marker=".", label="Quadratic"); \
            sdr.plot.time_domain(y3, sample_rate=rate, marker=".", label="Cubic"); \
            sdr.plot.time_domain(y4, sample_rate=rate, marker=".", label="Quartic"); \
            plt.xlim(1.5, 3.5); \
            plt.ylim(-1.0, -0.2); \
            plt.title("Comparison of Farrow Resamplers");

        Run a Farrow resampler with quartic polynomial order in streaming mode.

        .. ipython:: python

            x = np.cos(2 * np.pi / 5.179 * np.arange(40))
            farrow4 = sdr.FarrowResampler(4, streaming=True)

            y1 = farrow4(x[0:10], rate); \
            y2 = farrow4(x[10:20], rate); \
            y3 = farrow4(x[20:30], rate); \
            y4 = farrow4(x[30:40], rate); \
            y5 = farrow4.flush(rate); \
            y = np.concatenate((y1, y2, y3, y4, y5))

            @savefig sdr_FarrowResampler_6.png
            plt.figure(); \
            sdr.plot.time_domain(x, sample_rate=1, marker="o", label="Input"); \
            sdr.plot.time_domain(y, sample_rate=rate, offset=-farrow4.delay, marker=".", label="Quartic concatenated"); \
            plt.title("Quartic Farrow Resampler Concatenated Outputs");

        See the :ref:`farrow-arbitrary-resampler` example.

    Group:
        dsp-arbitrary-resampling
    """

    def __init__(self, order: int = 3, streaming: bool = False):
        """
        Creates a new Farrow arbitrary resampler.

        Arguments:
            order: The order of the piecewise polynomial, which must be between 1 and 4.
            streaming: Indicates whether to use streaming mode. In streaming mode, previous inputs are
                preserved between calls to :meth:`~FarrowResampler.__call__()`.

        Examples:
            See the :ref:`farrow-arbitrary-resampler` example.
        """
        self._order = verify_scalar(order, int=True, inclusive_min=1, inclusive_max=4)
        self._streaming = verify_bool(streaming)
        self._state: npt.NDArray  # FIR filter state. Will be updated in reset().
        self._mu_next: float  # The next fractional sample delay value

        if self.order == 1:
            self._delay = 1
            self._taps = np.array(
                [
                    [-1, 1],
                    [1, 0],
                ]
            ).T
        elif self.order == 2:
            self._delay = 2
            self._taps = np.array(
                [
                    [1 / 2, -1 / 2, 0],
                    [-1, 0, 1],
                    [1 / 2, 1 / 2, 0],
                ]
            ).T
        elif self.order == 3:
            self._delay = 2
            self._taps = np.array(
                [
                    [-1 / 6, 1 / 2, -1 / 3, 0],
                    [1 / 2, -1, -1 / 2, 1],
                    [-1 / 2, 1 / 2, 1, 0],
                    [1 / 6, 0, -1 / 6, 0],
                ]
            ).T
        elif self.order == 4:
            self._delay = 3
            self._taps = np.array(
                [
                    [1 / 24, -1 / 12, -1 / 24, 1 / 12, 0],
                    [-1 / 6, 1 / 6, 2 / 3, -2 / 3, 0],
                    [1 / 4, 0, -5 / 4, 0, 1],
                    [-1 / 6, -1 / 6, 2 / 3, 2 / 3, 0],
                    [1 / 24, 1 / 12, -1 / 24, -1 / 12, 0],
                ]
            ).T

        self.reset()

    ##############################################################################
    # Special methods
    ##############################################################################

    def __call__(self, x: npt.ArrayLike, rate: float) -> npt.NDArray:
        r"""
        Resamples the input signal $x[n]$ by the given arbitrary rate $r$.

        Arguments:
            x: The input signal $x[n] = x(n T_s)$.
            rate: The resampling rate $r$.

        Returns:
            The resampled signal $y[n] = x(n T_s / r)$.

        Examples:
            See the :ref:`farrow-arbitrary-resampler` example.
        """
        x = verify_arraylike(x, complex=True, atleast_1d=True, ndim=1)
        verify_scalar(rate, float=True, positive=True)

        if self.streaming:
            # Prepend previous inputs from last streaming call
            x_pad = np.concatenate((self._state, x))

            # Compute the FIR filter outputs for the entire input signal
            ys = []
            for i in range(self.order + 1):
                yi = scipy.signal.convolve(x_pad, self._taps[i, :], mode="full")
                ys.append(yi)

            # Compute the fractional sample indices for each output sample. We want to step from mu_next to
            # y0.size in steps of 1 / rate.
            frac_idxs = np.arange(
                self._mu_next,
                x_pad.size,
                1 / rate,
            )

            # Store the previous inputs and next fractional sample index for the next call to __call__()
            self._state = x_pad[-(self._taps.shape[1] - 1) :]
            self._mu_next = (frac_idxs[-1] + 1 / rate) - x_pad.size + (self._taps.shape[1] - 1)
        else:
            # Compute the FIR filter outputs for the entire input signal
            ys = []
            for i in range(self.order + 1):
                yi = scipy.signal.convolve(x, self._taps[i, :], mode="full")
                ys.append(yi)

            # Compute the fractional sample indices for each output sample
            frac_idxs = np.arange(
                self._delay,  # Account for filter delay
                self._delay + x.size - 1,
                1 / rate,
            )

        # Convert the fractional indices to integer indices and fractional indices
        int_idxs = (frac_idxs // 1.0).astype(int)
        mu = frac_idxs - int_idxs
        mu *= -1  # TODO: Why is this the case?

        # Interpolate the output samples using the Horner method
        y = ys[0][int_idxs]
        for i in range(1, self.order + 1):
            y += mu * y + ys[i][int_idxs]

        return convert_output(y)

    ##############################################################################
    # Streaming mode
    ##############################################################################

    def reset(self, state: npt.ArrayLike | None = None):
        """
        Resets the filter state and fractional sample index.

        Arguments:
            state: The filter state to reset to. The state vector should equal the previous three
                inputs. If `None`, the filter state will be reset to zero.

        Examples:
            See the :ref:`farrow-arbitrary-resampler` example.

        Group:
            Streaming mode only
        """
        if state is None:
            self._state = np.array([])
        else:
            self._state = verify_arraylike(state, complex=True, size=self._taps.shape[1] - 1)

        # Initial fractional sample delay accounts for filter delay
        self._mu_next = 0

    def flush(self, rate: float) -> npt.NDArray:
        """
        Flushes the filter state by passing zeros through the filter. Only useful when using streaming mode.

        Arguments:
            rate: The resampling rate $r$.

        Returns:
            The remaining resampled signal $y[n]$.

        Examples:
            See the :ref:`farrow-arbitrary-resampler` example.

        Group:
            Streaming mode only
        """
        verify_scalar(rate, float=True, positive=True)

        x = np.zeros_like(self.state)
        y = self(x, rate)

        return y

    @property
    def streaming(self) -> bool:
        """
        Indicates whether the filter is in streaming mode.

        In streaming mode, the filter state is preserved between calls to :meth:`~FarrowResampler.__call__()`.

        Examples:
            See the :ref:`farrow-arbitrary-resampler` example.

        Group:
            Streaming mode only
        """
        return self._streaming

    @property
    def state(self) -> npt.NDArray:
        """
        The filter state consisting of the previous $N$ inputs.

        Examples:
            See the :ref:`farrow-arbitrary-resampler` example.

        Group:
            Streaming mode only
        """
        return self._state

    ##############################################################################
    # Properties
    ##############################################################################

    @property
    def order(self) -> int:
        """
        The order of the piecewise polynomial.

        Examples:
            See the :ref:`farrow-arbitrary-resampler` example.
        """
        return self._order

    @property
    def taps(self) -> npt.NDArray:
        """
        The Farrow filter taps.

        Examples:
            See the :ref:`farrow-arbitrary-resampler` example.
        """
        return self._taps

    @property
    def delay(self) -> int:
        r"""
        The delay $d$ of the Farrow FIR filters in samples.

        Output sample $d \cdot r$, corresponds to the first input sample, where $r$ is the current resampling rate.

        Examples:
            See the :ref:`farrow-arbitrary-resampler` example.
        """
        return self._delay



"""
A module for signal amplitude synchronization.
"""

from __future__ import annotations

import numba
import numpy as np
import numpy.typing as npt

from .._helper import convert_output, export, verify_arraylike, verify_bool, verify_scalar


@export
class AGC:
    r"""
    Implements an automatic gain controller (AGC).

    Notes:
        .. code-block:: text
            :caption: Automatic Gain Control Block Diagram

            x[n] -->X---------------------------------------------------------------------+--> y[n]
                    ^                                                                     |
                    |                                                                     |
                +--------+      +------+                     -1  +--------+   +-------+   |
                | exp(.) |<--+--| z^-1 |<--@<---X<--------@<-----| log(.) |<--|  |.|  |<--+
                +--------+   |  +------+   ^   alpha    log(R)   +--------+   +-------+
                             |             |  or beta
                             +-------------+

            x[n] = Input signal
            y[n] = Output signal
            alpha = Attack rate
            beta = Decay rate
            R = Reference magnitude
            @ = Adder
            X = Multiplier

    References:
        - Michael Rice, *Digital Communications: A Discrete-Time Approach*, Section 9.5.
        - `Qasim Chaudhari, How Automatic Gain Control (AGC) Works.
          <https://wirelesspi.com/how-automatic-gain-control-agc-works/>`_

    Examples:
        Create an example received signal with two bursty signals surrounded by noise.

        .. ipython:: python

            x = np.exp(1j * 2 * np.pi * np.arange(5000) / 100); \
            x[0:1000] *= 0; \
            x[1000:2000] *= 10; \
            x[2000:3000] *= 0; \
            x[3000:4000] *= 0.1; \
            x[4000:5000] *= 0

            x += 0.001 * (np.random.randn(x.size) + 1j * np.random.randn(x.size))

            @savefig sdr_AGC_1.png
            plt.figure(); \
            sdr.plot.time_domain(x); \
            plt.title("Input signal");

        Create an AGC with an attack rate of $\alpha = 0.5$ and a decay rate of $\beta = 0.01$.
        Notice that over time the noise is amplified (according to the decay rate). Also notice that when the signal
        of interest appears the AGC gain is quickly decreased (according to the attack rate).

        .. ipython:: python

            agc = sdr.AGC(0.5, 0.01)
            y = agc(x)

            @savefig sdr_AGC_2.png
            plt.figure(); \
            sdr.plot.time_domain(y); \
            plt.ylim(-1.5, 1.5); \
            plt.title("Output signal");

    Group:
        synchronization-amplitude
    """

    def __init__(
        self,
        attack: float,
        decay: float,
        reference: float = 1.0,
        streaming: bool = False,
    ):
        r"""
        Creates an automatic gain controller (AGC).

        Arguments:
            attack: The attack rate $\alpha$. The attack rate is meant to attenuate strong signals.
                After $n_0 \approx 1 / \alpha$ samples the error will reduce to $1 / e$ of its original value.
            decay: The decay rate $\beta$. The decay rate is meant to amplify weak signals.
                After $n_0 \approx 1 / \beta$ samples the error will reduce to $1 / e$ of its original value.
            reference: The desired output magnitude.
            streaming: Indicates whether the AGC operates in streaming mode. In streaming mode, the gain is
                preserved between calls to :meth:`~AGC.__call__()`.
        """
        self.attack = attack  # Uses the property setter
        self.decay = decay  # Uses the property setter
        self.reference = reference  # Uses the property setter
        self._streaming = verify_bool(streaming)

        self._gain: float  # The linear gain state. Will be updated in reset().
        self.reset()

    def __call__(self, x: npt.ArrayLike) -> npt.NDArray[np.complex128]:
        """
        Performs automatic gain control on the input signal.

        Arguments:
            x: The input signal $x[n]$.

        Returns:
            The output signal $y[n]$.
        """
        x = verify_arraylike(x, complex=True, atleast_1d=True, ndim=1)

        if not self.streaming:
            self.reset()

        y, self._gain = _numba_agc_loop(x, self.attack, self.decay, self.reference, self._gain)

        return convert_output(y)

    ##############################################################################
    # Streaming mode
    ##############################################################################

    def reset(self, gain: float = 1.0):
        """
        Resets the AGC gain. Only useful when using streaming mode.

        Arguments:
            gain: The initial linear gain of the AGC.

        Group:
            Streaming mode only
        """
        self._gain = verify_scalar(gain, float=True, positive=True)

    @property
    def streaming(self) -> bool:
        """
        Indicates whether the AGC is in streaming mode.

        In streaming mode, the AGC gain is preserved between calls to :meth:`~AGC.__call__()`.

        Group:
            Streaming mode only
        """
        return self._streaming

    @property
    def gain(self) -> float:
        """
        The current linear gain.

        Group:
            Streaming mode only
        """
        return self._gain

    ##############################################################################
    # Properties
    ##############################################################################

    @property
    def attack(self) -> float:
        r"""
        (Settable) The attack rate $\alpha$.

        The attack rate is meant to attenuate strong signals. After $n_0 \approx 1 / \alpha$ samples the error will
        reduce to $1 / e$ of its original value.
        """
        return self._attack

    @attack.setter
    def attack(self, attack: float):
        self._attack = verify_scalar(attack, float=True, inclusive_min=0, inclusive_max=1)

    @property
    def decay(self) -> float:
        r"""
        (Settable) The decay rate $\beta$.

        The decay rate is meant to amplify weak signals. After $n_0 \approx 1 / \beta$ samples the error will reduce
        to $1 / e$ of its original value.
        """
        return self._decay

    @decay.setter
    def decay(self, decay: float):
        self._decay = verify_scalar(decay, float=True, inclusive_min=0, inclusive_max=1)

    @property
    def reference(self) -> float:
        """
        (Settable) The desired output magnitude.
        """
        return self._reference

    @reference.setter
    def reference(self, reference: float):
        self._reference = verify_scalar(reference, float=True, positive=True)


@numba.jit(nopython=True, cache=True)
def _numba_agc_loop(
    x: npt.NDArray[np.complex128],
    attack: float,
    decay: float,
    reference: float,
    gain: float,
) -> tuple[npt.NDArray[np.complex128], float]:
    y = np.zeros_like(x)
    reference = np.log(reference)
    gain = np.log(gain)  # Convert to log domain

    for i in range(x.size):
        y[i] = x[i] * np.exp(gain)

        y_mag = np.abs(y[i])
        y_mag = max(y_mag, 1e-100)  # Don't let the magnitude go to zero
        y_mag = np.log(y_mag)

        error = reference - y_mag
        if error < 0:
            gain += attack * error
        else:
            gain += decay * error

    gain = np.exp(gain)  # Convert back to linear domain

    return y, gain



import numpy as np

import sdr


def test_limit_cases():
    assert sdr.bsc_capacity(0) == 1
    assert sdr.bsc_capacity(1) == 1


def test_outputs():
    C = sdr.bsc_capacity([0, 0.5, 1])
    assert isinstance(C, np.ndarray)
    assert np.array_equal(C, [1, 0, 1])

    C = sdr.bsc_capacity(0.5)
    assert isinstance(C, float)
    assert C == 0



"""
A module for interpolating finite impulse response (FIR) filters.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt
import scipy.signal
from typing_extensions import Literal

from .._helper import convert_output, export, verify_arraylike, verify_bool, verify_literal, verify_scalar
from ._design_multirate import (
    multirate_fir,
    multirate_fir_linear,
    multirate_fir_linear_matlab,
    multirate_fir_zoh,
    polyphase_decompose,
)
from ._fir import FIR


def _polyphase_input_hold(
    x: npt.NDArray,
    state: npt.NDArray,
    H: npt.NDArray,
    mode: Literal["rate", "full"],
    streaming: bool,
) -> tuple[npt.NDArray, npt.NDArray]:
    B, M = H.shape  # Number of polyphase branches, polyphase filter length
    dtype = np.result_type(x, H)

    if streaming:
        x_pad = np.concatenate((state, x))  # Prepend previous inputs from last __call__() call

        Y = np.zeros((B, x.size), dtype=dtype)
        for i in range(B):
            Y[i] = scipy.signal.convolve(x_pad, H[i], mode="valid")

        state = x_pad[-(M - 1) :]
    else:
        if mode == "full":
            Y = np.zeros((B, x.size + M - 1), dtype=dtype)
            for i in range(B):
                Y[i] = scipy.signal.convolve(x, H[i], mode="full")
        elif mode == "rate":
            Y = np.zeros((B, x.size), dtype=dtype)
            for i in range(B):
                corr = scipy.signal.convolve(x, H[i], mode="full")
                Y[i] = corr[M // 2 : M // 2 + x.size]
        else:
            raise ValueError(f"Argument 'mode' must be 'rate' or 'full', not {mode!r}.")

    return Y, state


def _polyphase_input_commutate(
    x: npt.NDArray,
    state: npt.NDArray,
    H: npt.NDArray,
    mode: Literal["rate", "full"],
    bottom_to_top: bool,
    streaming: bool,
) -> tuple[npt.NDArray, npt.NDArray]:
    B, M = H.shape  # Number of polyphase branches, polyphase filter length
    dtype = np.result_type(x, H)

    if streaming:
        x_pad = np.concatenate((state, x))  # Prepend previous inputs from last __call__() call
        X_cols = x_pad.size // B
        X_pad = x_pad[0 : X_cols * B].reshape(X_cols, B).T  # Commutate across polyphase filters
        if bottom_to_top:
            X_pad = np.flipud(X_pad)  # Commutate from bottom to top

        Y = np.zeros((B, X_cols - M + 1), dtype=dtype)
        for i in range(B):
            Y[i] = scipy.signal.convolve(X_pad[i], H[i], mode="valid")

        state = x_pad[-(B * M - 1) :]
    else:
        # Prepend zeros to so the first sample is alone in the first commutated column
        x_pad = np.insert(x, 0, np.zeros(B - 1))
        # Append zeros to input signal to distribute evenly across the B branches
        x_pad = np.append(x_pad, np.zeros(B - (x_pad.size % B), dtype=dtype))
        X_cols = x_pad.size // B
        X_pad = x_pad.reshape(X_cols, B).T  # Commutate across polyphase filters
        if bottom_to_top:
            X_pad = np.flipud(X_pad)  # Commutate from bottom to top

        if mode == "full":
            Y = np.zeros((B, X_cols + M - 1), dtype=dtype)
            for i in range(B):
                Y[i] = scipy.signal.convolve(X_pad[i], H[i], mode="full")
        elif mode == "rate":
            Y = np.zeros((B, X_cols), dtype=dtype)
            for i in range(B):
                corr = scipy.signal.convolve(X_pad[i], H[i], mode="full")
                Y[i] = corr[M // 2 : M // 2 + X_cols]
        else:
            raise ValueError(f"Argument 'mode' must be 'rate' or 'full', not {mode!r}.")

    return Y, state


@export
class PolyphaseFIR(FIR):
    r"""
    Implements a generic polyphase FIR filter.

    Notes:
        The polyphase feedforward taps $h_i[n]$ are related to the prototype feedforward taps $h[n]$, given the
        number of polyphase branches $B$, by

        $$h_i[j] = h[i + j B] .$$

        The input signal $x[n]$ can be passed to each polyphase partition (used in interpolation) or commutated from
        bottom to top (used in decimation).

        .. md-tab-set::

            .. md-tab-item:: `input="hold"`

                .. code-block:: text
                   :caption: Polyphase FIR Filter with Input Hold

                                          +------------------------+
                                      +-->| h[0], h[3], h[6], h[9] |
                                      |   +------------------------+
                                      |   +------------------------+
                    ..., x[1], x[0] --+-->| h[1], h[4], h[7], 0    |
                                      |   +------------------------+
                                      |   +------------------------+
                                      +-->| h[2], h[5], h[8], 0    |
                                          +------------------------+

            .. md-tab-item:: `input="top-to-bottom"`

                .. code-block:: text
                   :caption: Polyphase FIR Filter with Input Commutation (Top to Bottom)

                                             +------------------------+
                    ..., x[4], x[1], 0    -->| h[0], h[3], h[6], h[9] |
                                             +------------------------+
                                             +------------------------+
                    ..., x[5], x[2], 0    -->| h[1], h[4], h[7], 0    |
                                             +------------------------+
                                             +------------------------+
                    ..., x[6], x[3], x[0] -->| h[2], h[5], h[8], 0    |
                                             +------------------------+

            .. md-tab-item:: `input="bottom-to-top"`

                .. code-block:: text
                   :caption: Polyphase FIR Filter with Input Commutation (Bottom to Top)

                                             +------------------------+
                    ..., x[6], x[3], x[0] -->| h[0], h[3], h[6], h[9] |
                                             +------------------------+
                                             +------------------------+
                    ..., x[5], x[2], 0    -->| h[1], h[4], h[7], 0    |
                                             +------------------------+
                                             +------------------------+
                    ..., x[4], x[1], 0    -->| h[2], h[5], h[8], 0    |
                                             +------------------------+

        The output of each polyphase partition can be summed to produce the output signal $y[n]$ (used in decimation),
        commutated from top to bottom (used in interpolation), or taken as parallel outputs $y_i[n]$ (used in
        channelization).

        .. md-tab-set::

            .. md-tab-item:: `output="sum"`

                .. code-block:: text
                   :caption: Polyphase FIR Filter with Output Summation

                    +------------------------+
                    | h[0], h[3], h[6], h[9] |--+
                    +------------------------+  |
                    +------------------------+  v
                    | h[1], h[4], h[7], 0    |--@--> ..., y[1], y[0]
                    +------------------------+  ^
                    +------------------------+  |
                    | h[2], h[5], h[8], 0    |--+
                    +------------------------+

            .. md-tab-item:: `output="top-to-bottom"`

                .. code-block:: text
                   :caption: Polyphase FIR Filter with Output Commutation (Top to Bottom)

                    +------------------------+
                    | h[0], h[3], h[6], h[9] |--> ..., y[3], y[0]
                    +------------------------+
                    +------------------------+
                    | h[1], h[4], h[7], 0    |--> ..., y[4], y[1]
                    +------------------------+
                    +------------------------+
                    | h[2], h[5], h[8], 0    |--> ..., y[5], y[2]
                    +------------------------+

            .. md-tab-item:: `output="bottom-to-top"`

                .. code-block:: text
                   :caption: Polyphase FIR Filter with Output Commutation (Bottom to Top)

                    +------------------------+
                    | h[0], h[3], h[6], h[9] |--> ..., y[5], y[2]
                    +------------------------+
                    +------------------------+
                    | h[1], h[4], h[7], 0    |--> ..., y[4], y[1]
                    +------------------------+
                    +------------------------+
                    | h[2], h[5], h[8], 0    |--> ..., y[3], y[0]
                    +------------------------+

            .. md-tab-item:: `output="all"`

                .. code-block:: text
                   :caption: Polyphase FIR Filter with Parallel Outputs

                    +------------------------+
                    | h[0], h[3], h[6], h[9] |--> ..., y[0,1], y[0,0]
                    +------------------------+
                    +------------------------+
                    | h[1], h[4], h[7], 0    |--> ..., y[1,1], y[1,0]
                    +------------------------+
                    +------------------------+
                    | h[2], h[5], h[8], 0    |--> ..., y[2,1], y[2,0]
                    +------------------------+

    Group:
        dsp-polyphase-filtering
    """

    def __init__(
        self,
        branches: int,
        taps: npt.ArrayLike,
        input: Literal["hold", "top-to-bottom", "bottom-to-top"] = "hold",
        output: Literal["sum", "top-to-bottom", "bottom-to-top", "all"] = "sum",
        streaming: bool = False,
    ):
        r"""
        Creates a polyphase FIR filter.

        Arguments:
            branches: The number of polyphase branches $B$.
            taps: The prototype filter feedforward coefficients $h[n]$.
            input: The input connection method.

                - `"hold"`: The input signal $x[n]$ is passed to each polyphase partition (used in interpolation).
                - `"top-to-bottom"`: The input signal $x[n]$ is commutated across the polyphase partitions from
                  top to bottom.
                - `"bottom-to-top"`: The input signal $x[n]$ is commutated across the polyphase partitions from
                  bottom to top (used in decimation).

            output: The output connection method.

                - `"sum"`: The output of each polyphase partition is summed to produce the output signal $y[n]$ (used
                  in decimation).
                - `"top-to-bottom"`: The output of each polyphase partition is commutated from top to bottom (used in
                  interpolation).
                - `"bottom-to-top"`: The output of each polyphase partition is commutated from bottom to top.
                - `"all"`: The outputs of each polyphase partition are used in parallel (used in channelization).

            streaming: Indicates whether to use streaming mode. In streaming mode, previous inputs are
                preserved between calls to :meth:`~PolyphaseFIR.__call__()`.
        """
        self._branches = verify_scalar(branches, int=True, positive=True)
        self._taps = verify_arraylike(taps, ndim=1)
        self._input = verify_literal(input, ["hold", "top-to-bottom", "bottom-to-top"])
        self._output = verify_literal(output, ["sum", "top-to-bottom", "bottom-to-top", "all"])
        verify_bool(streaming)

        self._polyphase_taps = polyphase_decompose(self.branches, self.taps)

        # Determine the effective interpolation and decimation rates based on the input and output connection types
        self._interpolation = 1
        self._decimation = 1
        if self.input in ["top-to-bottom", "bottom-to-top"]:
            self._decimation *= self.branches
        if self.output in ["top-to-bottom", "bottom-to-top"]:
            self._interpolation *= self.branches

        super().__init__(taps, streaming=streaming)

    ##############################################################################
    # Special methods
    ##############################################################################

    def __call__(
        self,
        x: npt.ArrayLike,
        mode: Literal["rate", "full"] = "rate",
    ) -> npt.NDArray:
        r"""
        Filters the input signal $x[n]$ with the polyphase FIR filter.

        Arguments:
            x: The input signal $x[n]$ with sample rate $f_s$ and length $L$.
            mode: The non-streaming convolution mode.

                - `"rate"`: The output signal $y[n]$ has length $L \cdot r$, proportional to the resampling rate
                  $r$. Output sample 0 aligns with input sample 0.
                - `"full"`: The full convolution is performed. The output signal $y[n]$ has length $(L + N) \cdot r$,
                  where $N$ is the order of the prototype filter. Output sample :obj:`~PolyphaseFIR.delay` aligns
                  with input sample 0.

                In streaming mode, the `"full"` convolution is performed. However, for each $L$ input samples
                only $L \cdot r$ output samples are produced per call. A final call to :meth:`~PolyphaseFIR.flush()`
                is required to flush the filter state.

        Returns:
            The filtered signal $y[n]$ with sample rate $f_s \cdot r$. The output length is dictated by
            the `mode` argument.
        """
        x = verify_arraylike(x, atleast_1d=True, ndim=1)
        verify_literal(mode, ["rate", "full"])

        if self.input == "hold":
            Y, self._state = _polyphase_input_hold(x, self._state, self.polyphase_taps, mode, self.streaming)
        elif self.input == "top-to-bottom":
            Y, self._state = _polyphase_input_commutate(
                x, self._state, self.polyphase_taps, mode, False, self.streaming
            )
        elif self.input == "bottom-to-top":
            Y, self._state = _polyphase_input_commutate(x, self._state, self.polyphase_taps, mode, True, self.streaming)

        if self.output == "sum":
            y = np.sum(Y, axis=0)
        elif self.output == "bottom-to-top":
            y = np.flipud(Y).T.flatten()
        elif self.output == "top-to-bottom":
            y = Y.T.flatten()
        elif self.output == "all":
            y = Y

        return convert_output(y)

    def __repr__(self) -> str:
        h_str = np.array2string(self.taps, max_line_width=int(1e6), separator=", ", suppress_small=True)
        return f"sdr.{type(self).__name__}({self.branches}, {h_str}, streaming={self.streaming})"

    def __str__(self) -> str:
        string = f"sdr.{type(self).__name__}:"
        string += f"\n  order: {self.order}"
        string += f"\n  branches: {self.branches}"
        string += f"\n  rate: {self.interpolation} / {self.decimation}"
        string += f"\n  polyphase_taps: {self.polyphase_taps.shape} shape"
        string += f"\n  delay: {self.delay}"
        string += f"\n  streaming: {self.streaming}"
        return string

    ##############################################################################
    # Streaming mode
    ##############################################################################

    def reset(self):
        self._state = np.zeros(self.polyphase_taps.shape[1] - 1)

    ##############################################################################
    # Properties
    ##############################################################################

    @property
    def branches(self) -> int:
        """
        The number of polyphase branches $B$.
        """
        return self._branches

    @property
    def taps(self) -> npt.NDArray:
        """
        The prototype feedforward taps $h[n]$.

        Notes:
            The prototype feedforward taps $h[n]$ are the feedforward taps of the FIR filter before
            polyphase decomposition. The polyphase feedforward taps $h_i[n]$ are the feedforward taps
            of the FIR filter after polyphase decomposition.

            The polyphase feedforward taps $h_i[n]$ are related to the prototype feedforward taps $h[n]$, given the
            number of polyphase branches $B$, by

            $$h_i[j] = h[i + j B] .$$

        Examples:
            .. ipython:: python

                fir = sdr.PolyphaseFIR(3, np.arange(10))
                fir.taps
                fir.polyphase_taps
        """
        return self._taps

    @property
    def polyphase_taps(self) -> npt.NDArray:
        """
        The polyphase feedforward taps $h_i[n]$.

        Notes:
            The prototype feedforward taps $h[n]$ are the feedforward taps of the FIR filter before
            polyphase decomposition. The polyphase feedforward taps $h_i[n]$ are the feedforward taps
            of the FIR filter after polyphase decomposition.

            The polyphase feedforward taps $h_i[n]$ are related to the prototype feedforward taps $h[n]$, given the
            number of polyphase branches $B$, by

            $$h_i[j] = h[i + j B] .$$

        Examples:
            .. ipython:: python

                fir = sdr.PolyphaseFIR(3, np.arange(10))
                fir.taps
                fir.polyphase_taps
        """
        return self._polyphase_taps

    @property
    def order(self) -> int:
        """
        The order $N = (M + 1)B - 1$ of the FIR prototype filter $h[n]$.

        Examples:
            .. ipython:: python

                fir = sdr.PolyphaseFIR(3, np.arange(10))
                fir.taps
                fir.polyphase_taps
                fir.order
                fir.polyphase_order
        """
        return self.taps.size - 1

    @property
    def polyphase_order(self) -> int:
        """
        The order $M = (N + 1)/B - 1$ of each FIR polyphase filter $h_i[n]$.

        Examples:
            .. ipython:: python

                fir = sdr.PolyphaseFIR(3, np.arange(10))
                fir.taps
                fir.polyphase_taps
                fir.order
                fir.polyphase_order
        """
        return self.polyphase_taps.shape[1] - 1

    @property
    def input(self) -> Literal["hold", "top-to-bottom", "bottom-to-top"]:
        """
        The input connection method.

        Notes:
            - `"hold"`: The input signal $x[n]$ is passed to each polyphase partition (used in interpolation).
            - `"top-to-bottom"`: The input signal $x[n]$ is commutated from top to bottom.
            - `"bottom-to-top"`: The input signal $x[n]$ is commutated from bottom to top (used in decimation).
        """
        return self._input

    @property
    def output(self) -> Literal["sum", "top-to-bottom", "bottom-to-top", "all"]:
        """
        The output connection method.

        Notes:
            - `"sum"`: The output of each polyphase partition is summed to produce the output signal $y[n]$ (used in
              decimation).
            - `"top-to-bottom"`: The output of each polyphase partition is commutated from top to bottom (used in
              interpolation).
            - `"bottom-to-top"`: The output of each polyphase partition is commutated from bottom to top.
            - `"all"`: The outputs of each polyphase partition are used in parallel (used in channelization).
        """
        return self._output

    @property
    def interpolation(self) -> int:
        """
        The integer interpolation rate $P$.
        """
        return self._interpolation

    @property
    def decimation(self) -> int:
        """
        The integer decimation rate $Q$.
        """
        return self._decimation

    @property
    def rate(self) -> float:
        r"""
        The fractional resampling rate $r = P/Q$. The output sample rate is $f_{s,out} = f_{s,in} \cdot r$.
        """
        return self.interpolation / self.decimation

    @property
    def delay(self) -> int:
        """
        The delay of polyphase FIR filter in samples.

        The delay indicates the output sample index that corresponds to the first input sample.
        """
        return super().delay // self.decimation


@export
class Interpolator(PolyphaseFIR):
    r"""
    Implements a polyphase interpolating FIR filter.

    Notes:
        The polyphase interpolating filter is equivalent to first upsampling the input signal $x[n]$ by $P$
        (by inserting $P-1$ zeros between each sample) and then filtering the upsampled signal with the
        prototype FIR filter with feedforward coefficients $h[n]$.

        Instead, the polyphase interpolating filter first decomposes the prototype FIR filter into $P$ polyphase
        filters with feedforward coefficients $h_i[n]$. The polyphase filters are then applied to the
        input signal $x[n]$ in parallel. The output of the polyphase filters are then commutated to produce
        the output signal $y[n]$. This prevents the need to multiply with zeros in the upsampled input,
        as is needed in the first case.

        .. code-block:: text
           :caption: Polyphase 3x Interpolating FIR Filter Block Diagram

                                  +------------------------+
                              +-->| h[0], h[3], h[6], h[9] |--> ..., y[3], y[0]
                              |   +------------------------+
                              |   +------------------------+
            ..., x[1], x[0] --+-->| h[1], h[4], h[7], 0    |--> ..., y[4], y[1]
                              |   +------------------------+
                              |   +------------------------+
                              +-->| h[2], h[5], h[8], 0    |--> ..., y[5], y[2]
                                  +------------------------+

            Input Hold                                          Output Commutator
                                                                (top-to-bottom)

            x[n] = Input signal with sample rate fs
            y[n] = Output signal with sample rate fs * P
            h[n] = Prototype FIR filter

        The polyphase feedforward taps $h_i[n]$ are related to the prototype feedforward taps $h[n]$ by

        $$h_i[j] = h[i + j P] .$$

    References:
        - fred harris, *Multirate Signal Processing for Communication Systems*, Chapter 7: Resampling Filters.

    Examples:
        Create an input signal to interpolate.

        .. ipython:: python

            x = np.cos(np.pi / 4 * np.arange(40))

        Create a polyphase filter that interpolates by 7 using the Kaiser window method.

        .. ipython:: python

            fir = sdr.Interpolator(7); fir
            y = fir(x)

            @savefig sdr_Interpolator_1.png
            plt.figure(); \
            sdr.plot.time_domain(x, marker="o", label="Input"); \
            sdr.plot.time_domain(y, sample_rate=fir.rate, marker=".", label="Interpolated"); \
            plt.title("Interpolation by 7 with the Kaiser window method");

        Create a streaming polyphase filter that interpolates by 7 using the Kaiser window method. This filter
        preserves state between calls.

        .. ipython:: python

            fir = sdr.Interpolator(7, streaming=True); fir

            y1 = fir(x[0:10]); \
            y2 = fir(x[10:20]); \
            y3 = fir(x[20:30]); \
            y4 = fir(x[30:40]); \
            y5 = fir.flush()

            @savefig sdr_Interpolator_2.png
            plt.figure(); \
            sdr.plot.time_domain(x, marker="o", label="Input"); \
            sdr.plot.time_domain(y1, sample_rate=fir.rate, offset=-fir.delay/fir.rate + 0, marker=".", label="Interpolated $y_1[n]$"); \
            sdr.plot.time_domain(y2, sample_rate=fir.rate, offset=-fir.delay/fir.rate + 10, marker=".", label="Interpolated $y_2[n]$"); \
            sdr.plot.time_domain(y3, sample_rate=fir.rate, offset=-fir.delay/fir.rate + 20, marker=".", label="Interpolated $y_3[n]$"); \
            sdr.plot.time_domain(y4, sample_rate=fir.rate, offset=-fir.delay/fir.rate + 30, marker=".", label="Interpolated $y_4[n]$"); \
            sdr.plot.time_domain(y5, sample_rate=fir.rate, offset=-fir.delay/fir.rate + 40, marker=".", label="Interpolated $y_5[n]$"); \
            plt.title("Streaming interpolation by 7 with the Kaiser window method");

        Create a polyphase filter that interpolates by 7 using linear method.

        .. ipython:: python

            fir = sdr.Interpolator(7, "linear"); fir
            y = fir(x)

            @savefig sdr_Interpolator_3.png
            plt.figure(); \
            sdr.plot.time_domain(x, marker="o", label="Input"); \
            sdr.plot.time_domain(y, sample_rate=fir.rate, marker=".", label="Interpolated"); \
            plt.title("Interpolation by 7 with the linear method");

        Create a polyphase filter that interpolates by 7 using the zero-order hold method. It is recommended to use
        the `"full"` convolution mode. This way the first upsampled symbol has $r$ samples.

        .. ipython:: python

            fir = sdr.Interpolator(7, "zoh"); fir
            y = fir(x, mode="full")

            @savefig sdr_Interpolator_4.png
            plt.figure(); \
            sdr.plot.time_domain(x, marker="o", label="Input"); \
            sdr.plot.time_domain(y, sample_rate=fir.rate, offset=-fir.delay/fir.rate, marker=".", label="Interpolated"); \
            plt.title("Interpolation by 7 with the zero-order hold method");

    Group:
        dsp-polyphase-filtering
    """

    def __init__(
        self,
        interpolation: int,
        taps: Literal["kaiser", "linear", "linear-matlab", "zoh"] | npt.ArrayLike = "kaiser",
        polyphase_order: int = 23,
        atten: float = 80,
        streaming: bool = False,
    ):
        r"""
        Creates a polyphase FIR interpolating filter.

        Arguments:
            interpolation: The interpolation rate $P$.
            taps: The prototype filter design specification.

                - `"kaiser"`: The prototype filter is designed using :func:`sdr.multirate_fir`
                  with arguments `interpolation` and 1.
                - `"linear"`: The prototype filter is designed to linearly interpolate between samples.
                  The filter coefficients are a length-$2P$ linear ramp $\frac{1}{P} [0, ..., P-1, P, P-1, ..., 1]$.
                  The first output sample aligns with the first input sample.
                - `"linear-matlab"`: The prototype filter is designed to linearly interpolate between samples.
                  The filter coefficients are a length-$2P$ linear ramp $\frac{1}{P} [1, ..., P-1, P, P-1, ..., 0]$.
                  This is method MATLAB uses. The first output sample is advanced from the first input sample.
                - `"zoh"`: The prototype filter is designed to be a zero-order hold.
                  The filter coefficients are a length-$P$ array of ones.
                - `npt.ArrayLike`: The prototype filter feedforward coefficients $h[n]$.

            polyphase_order: The order of each polyphase filter. Must be odd, such that the filter lengths are even.
                Only used when `taps="kaiser"`.
            atten: The stopband attenuation $A_{\text{stop}}$ in dB. Only used when `taps="kaiser"`.
            streaming: Indicates whether to use streaming mode. In streaming mode, previous inputs are
                preserved between calls to :meth:`~Interpolator.__call__()`.
        """
        verify_scalar(interpolation, int=True, positive=True)
        verify_scalar(polyphase_order, int=True, positive=True)
        verify_scalar(atten, float=True, positive=True)
        verify_bool(streaming)

        self._method: Literal["kaiser", "linear", "linear-matlab", "zoh", "custom"]
        if not isinstance(taps, str):
            self._method = "custom"
            taps = verify_arraylike(taps, atleast_1d=True, ndim=1)
        else:
            self._method = verify_literal(taps, ["kaiser", "linear", "linear-matlab", "zoh"])
            if taps == "kaiser":
                taps = multirate_fir(interpolation, 1, polyphase_order, atten)
            elif taps == "linear":
                taps = multirate_fir_linear(interpolation)
            elif taps == "linear-matlab":
                taps = multirate_fir_linear_matlab(interpolation)
            elif taps == "zoh":
                taps = multirate_fir_zoh(interpolation)

        super().__init__(interpolation, taps, input="hold", output="top-to-bottom", streaming=streaming)

    ##############################################################################
    # Special methods
    ##############################################################################

    def __repr__(self) -> str:
        if self.method == "custom":
            h_str = np.array2string(self.taps, max_line_width=int(1e6), separator=", ", suppress_small=True)
        else:
            h_str = repr(self.method)
        return f"sdr.{type(self).__name__}({self.interpolation}, {h_str}, streaming={self.streaming})"

    def __str__(self) -> str:
        string = f"sdr.{type(self).__name__}:"
        string += f"\n  order: {self.order}"
        string += f"\n  rate: {self.interpolation} / {self.decimation}"
        string += f"\n  method: {self.method!r}"
        string += f"\n  polyphase_taps: {self.polyphase_taps.shape} shape"
        string += f"\n  delay: {self.delay}"
        string += f"\n  streaming: {self.streaming}"
        return string

    ##############################################################################
    # Properties
    ##############################################################################

    @property
    def method(self) -> Literal["kaiser", "linear", "linear-matlab", "zoh", "custom"]:
        """
        The method used to design the polyphase interpolating filter.
        """
        return self._method


@export
class Decimator(PolyphaseFIR):
    r"""
    Implements a polyphase decimating FIR filter.

    Notes:
        The polyphase decimating filter is equivalent to first filtering the input signal $x[n]$ with the
        prototype FIR filter with feedforward coefficients $h[n]$ and then downsampling the filtered signal
        by $Q$ (by discarding $Q-1$ samples every $Q$ samples).

        Instead, the polyphase decimating filter first decomposes the prototype FIR filter into $Q$ polyphase
        filters with feedforward coefficients $h_i[n]$. The polyphase filters are then applied to the
        commutated input signal $x[n]$ in parallel. The outputs of the polyphase filters are then summed.
        This prevents the need to compute outputs that will be discarded, as is done in the first case.

        .. code-block:: text
           :caption: Polyphase 3x Decimating FIR Filter Block Diagram

                                     +------------------------+
            ..., x[6], x[3], x[0] -->| h[0], h[3], h[6], h[9] |---+
                                     +------------------------+   |
                                     +------------------------+   v
            ..., x[5], x[2], 0    -->| h[1], h[4], h[7], 0    |-->@--> ..., y[1], y[0]
                                     +------------------------+   ^
                                     +------------------------+   |
            ..., x[4], x[1], 0    -->| h[2], h[5], h[8], 0    |---+
                                     +------------------------+

            Input Commutator                                           Output Summation
            (bottom-to-top)

            x[n] = Input signal with sample rate fs
            y[n] = Output signal with sample rate fs / Q
            h[n] = Prototype FIR filter
            @ = Adder

        The polyphase feedforward taps $h_i[n]$ are related to the prototype feedforward taps $h[n]$ by

        $$h_i[j] = h[i + j Q] .$$

    References:
        - fred harris, *Multirate Signal Processing for Communication Systems*, Chapter 7: Resampling Filters.

    Examples:
        Create an input signal to interpolate.

        .. ipython:: python

            x = np.cos(np.pi / 64 * np.arange(280))

        Create a polyphase filter that decimates by 7 using the Kaiser window method.

        .. ipython:: python

            fir = sdr.Decimator(7); fir
            y = fir(x)

            @savefig sdr_Decimator_1.png
            plt.figure(); \
            sdr.plot.time_domain(x, marker=".", label="Input"); \
            sdr.plot.time_domain(y, sample_rate=fir.rate, marker="o", label="Decimated"); \
            plt.title("Decimation by 7 with the Kaiser window method");

        Create a streaming polyphase filter that decimates by 7 using the Kaiser window method. This filter
        preserves state between calls.

        .. ipython:: python

            fir = sdr.Decimator(7, streaming=True); fir

            y1 = fir(x[0:70]); \
            y2 = fir(x[70:140]); \
            y3 = fir(x[140:210]); \
            y4 = fir(x[210:280]); \
            y5 = fir.flush()

            @savefig sdr_Decimator_2.png
            plt.figure(); \
            sdr.plot.time_domain(x, marker=".", label="Input"); \
            sdr.plot.time_domain(y1, sample_rate=fir.rate, offset=-fir.delay/fir.rate + 0, marker="o", label="Decimated $y_1[n]$"); \
            sdr.plot.time_domain(y2, sample_rate=fir.rate, offset=-fir.delay/fir.rate + 70, marker="o", label="Decimated $y_2[n]$"); \
            sdr.plot.time_domain(y3, sample_rate=fir.rate, offset=-fir.delay/fir.rate + 140, marker="o", label="Decimated $y_3[n]$"); \
            sdr.plot.time_domain(y4, sample_rate=fir.rate, offset=-fir.delay/fir.rate + 210, marker="o", label="Decimated $y_4[n]$"); \
            sdr.plot.time_domain(y5, sample_rate=fir.rate, offset=-fir.delay/fir.rate + 280, marker="o", label="Decimated $y_5[n]$"); \
            plt.title("Streaming decimation by 7 with the Kaiser window method");

    Group:
        dsp-polyphase-filtering
    """

    def __init__(
        self,
        decimation: int,
        taps: Literal["kaiser"] | npt.ArrayLike = "kaiser",
        polyphase_order: int = 23,
        atten: float = 80,
        streaming: bool = False,
    ):
        r"""
        Creates a polyphase FIR decimating filter.

        Arguments:
            decimation: The decimation rate $Q$.
            taps: The prototype filter design specification.

                - `"kaiser"`: The prototype filter is designed using :func:`sdr.multirate_fir`
                  with arguments 1 and `decimation`.
                - `npt.ArrayLike`: The prototype filter feedforward coefficients $h[n]$.

            polyphase_order: The order of each polyphase filter. Must be odd, such that the filter lengths are even.
                Only used when `taps="kaiser"`.
            atten: The stopband attenuation $A_{\text{stop}}$ in dB. Only used when `taps="kaiser"`.
            streaming: Indicates whether to use streaming mode. In streaming mode, previous inputs are
                preserved between calls to :meth:`~Decimator.__call__()`.
        """
        verify_scalar(decimation, int=True, positive=True)

        self._method: Literal["kaiser", "custom"]
        if not isinstance(taps, str):
            self._method = "custom"
            taps = verify_arraylike(taps, atleast_1d=True, ndim=1)
        else:
            self._method = verify_literal(taps, ["kaiser"])
            taps = multirate_fir(1, decimation, polyphase_order, atten)

        super().__init__(decimation, taps, input="bottom-to-top", output="sum", streaming=streaming)

    ##############################################################################
    # Special methods
    ##############################################################################

    def __repr__(self) -> str:
        if self.method == "custom":
            h_str = np.array2string(self.taps, max_line_width=int(1e6), separator=", ", suppress_small=True)
        else:
            h_str = repr(self.method)
        return f"sdr.{type(self).__name__}({self.decimation}, {h_str}, streaming={self.streaming})"

    def __str__(self) -> str:
        string = f"sdr.{type(self).__name__}:"
        string += f"\n  order: {self.order}"
        string += f"\n  rate: {self.interpolation} / {self.decimation}"
        string += f"\n  method: {self.method!r}"
        string += f"\n  polyphase_taps: {self.polyphase_taps.shape} shape"
        string += f"\n  delay: {self.delay}"
        string += f"\n  streaming: {self.streaming}"
        return string

    ##############################################################################
    # Streaming mode
    ##############################################################################

    def reset(self):
        self._state = np.zeros(self.polyphase_taps.size - 1)

    ##############################################################################
    # Properties
    ##############################################################################

    @property
    def method(self) -> Literal["kaiser", "custom"]:
        """
        The method used to design the polyphase decimating filter.
        """
        return self._method


@export
class Resampler(PolyphaseFIR):
    r"""
    Implements a polyphase rational resampling FIR filter.

    Notes:
        The polyphase rational resampling filter is equivalent to first upsampling the input signal $x[n]$ by $P$
        (by inserting $P-1$ zeros between each sample), filtering the upsampled signal with the prototype FIR filter
        with feedforward coefficients $h[n]$, and then downsampling the filtered signal by $Q$ (by discarding $Q-1$
        samples every $Q$ samples).

        Instead, the polyphase rational resampling filter first decomposes the prototype FIR filter into $P$ polyphase
        filters with feedforward coefficients $h_i[n]$. The polyphase filters are then applied to the
        input signal $x[n]$ in parallel. The output of the polyphase filters are then commutated by $Q$ to produce
        the output signal $y[n]$. This prevents the need to multiply with zeros in the upsampled input,
        as is needed in the first case.

        .. code-block:: text
           :caption: Polyphase 3/2 Resampling FIR Filter Block Diagram

                                  +------------------------+
                              +-->| h[0], h[3], h[6], h[9] |--> ..., ____, y[0]
                              |   +------------------------+
                              |   +------------------------+
            ..., x[1], x[0] --+-->| h[1], h[4], h[7], 0    |--> ..., y[2], ____
                              |   +------------------------+
                              |   +------------------------+
                              +-->| h[2], h[5], h[8], 0    |--> ..., ____, y[1]
                                  +------------------------+

            Input Hold                                          Output Commutator by 2
                                                                (top-to-bottom)

            x[n] = Input signal with sample rate fs
            y[n] = Output signal with sample rate fs * P / Q
            h[n] = Prototype FIR filter

        The polyphase feedforward taps $h_i[n]$ are related to the prototype feedforward taps $h[n]$ by

        $$h_i[j] = h[i + j P] .$$

        If the interpolation rate $P$ is 1, then the polyphase rational resampling filter is equivalent to the
        polyphase decimating filter. See :class:`~sdr.Decimator`.

    References:
        - fred harris, *Multirate Signal Processing for Communication Systems*, Chapter 7: Resampling Filters.

    Examples:
        Create an input signal to resample.

        .. ipython:: python

            x = np.cos(np.pi / 4 * np.arange(40))

        Create a polyphase filter that resamples by 7/3 using the Kaiser window method.

        .. ipython:: python

            fir = sdr.Resampler(7, 3); fir
            y = fir(x)

            @savefig sdr_Resampler_1.png
            plt.figure(); \
            sdr.plot.time_domain(x, marker="o", label="Input"); \
            sdr.plot.time_domain(y, sample_rate=fir.rate, marker=".", label="Resampled"); \
            plt.title("Resampling by 7/3 with the Kaiser window method");

        Create a streaming polyphase filter that resamples by 7/3 using the Kaiser window method. This filter
        preserves state between calls.

        .. ipython:: python

            fir = sdr.Resampler(7, 3, streaming=True); fir

            y1 = fir(x[0:10]); \
            y2 = fir(x[10:20]); \
            y3 = fir(x[20:30]); \
            y4 = fir(x[30:40]); \
            y5 = fir.flush()

            @savefig sdr_Resampler_2.png
            plt.figure(); \
            sdr.plot.time_domain(x, marker="o", label="Input"); \
            sdr.plot.time_domain(y1, sample_rate=fir.rate, offset=-fir.delay/fir.rate + 0, marker=".", label="Resampled $y_1[n]$"); \
            sdr.plot.time_domain(y2, sample_rate=fir.rate, offset=-fir.delay/fir.rate + 10, marker=".", label="Resampled $y_2[n]$"); \
            sdr.plot.time_domain(y3, sample_rate=fir.rate, offset=-fir.delay/fir.rate + 20, marker=".", label="Resampled $y_3[n]$"); \
            sdr.plot.time_domain(y4, sample_rate=fir.rate, offset=-fir.delay/fir.rate + 30, marker=".", label="Resampled $y_4[n]$"); \
            sdr.plot.time_domain(y5, sample_rate=fir.rate, offset=-fir.delay/fir.rate + 40, marker=".", label="Resampled $y_5[n]$"); \
            plt.title("Streaming resampling by 7/3 with the Kaiser window method");

        Create a polyphase filter that resamples by 5/7 using linear method.

        .. ipython:: python

            fir = sdr.Resampler(5, 7); fir
            y = fir(x)

            @savefig sdr_Resampler_3.png
            plt.figure(); \
            sdr.plot.time_domain(x, marker=".", label="Input"); \
            sdr.plot.time_domain(y, sample_rate=fir.rate, marker="o", label="Resampled"); \
            plt.title("Resampling by 5/7 with the Kaiser window method");

    Group:
        dsp-polyphase-filtering
    """

    def __init__(
        self,
        interpolation: int,
        decimation: int,
        taps: Literal["kaiser", "linear", "linear-matlab", "zoh"] | npt.ArrayLike = "kaiser",
        polyphase_order: int = 23,
        atten: float = 80,
        streaming: bool = False,
    ):
        r"""
        Creates a polyphase FIR rational resampling filter.

        Arguments:
            interpolation: The interpolation rate $P$.
            decimation: The decimation rate $Q$.
            taps: The prototype filter design specification.

                - `"kaiser"`: The prototype filter is designed using :func:`sdr.multirate_fir`
                  with arguments `interpolation` and `decimation`.
                - `"linear"`: The prototype filter is designed to linearly interpolate between samples.
                  The filter coefficients are a length-$2P$ linear ramp $\frac{1}{P} [0, ..., P-1, P, P-1, ..., 1]$.
                  The first output sample aligns with the first input sample.
                - `"linear-matlab"`: The prototype filter is designed to linearly interpolate between samples.
                  The filter coefficients are a length-$2P$ linear ramp $\frac{1}{P} [1, ..., P-1, P, P-1, ..., 0]$.
                  This is method MATLAB uses. The first output sample is advanced from the first input sample.
                - `"zoh"`: The prototype filter is designed to be a zero-order hold.
                  The filter coefficients are a length-$P$ array of ones.
                - `npt.ArrayLike`: The prototype filter feedforward coefficients $h[n]$.

            polyphase_order: The order of each polyphase filter. Must be odd, such that the filter lengths are even.
                Only used when `taps="kaiser"`.
            atten: The stopband attenuation $A_{\text{stop}}$ in dB. Only used when `taps="kaiser"`.
            streaming: Indicates whether to use streaming mode. In streaming mode, previous inputs are
                preserved between calls to :meth:`~Resampler.__call__()`.
        """
        verify_scalar(interpolation, int=True, positive=True)
        verify_scalar(decimation, int=True, positive=True)

        self._method: Literal["kaiser", "linear", "linear-matlab", "zoh", "custom"]
        if not isinstance(taps, str):
            self._method = "custom"
            taps = verify_arraylike(taps, atleast_1d=True, ndim=1)
        else:
            self._method = verify_literal(taps, ["kaiser", "linear", "linear-matlab", "zoh"])
            if taps == "kaiser":
                taps = multirate_fir(interpolation, decimation, polyphase_order, atten)
            else:
                verify_scalar(interpolation, int=True, exclusive_min=1)
                if taps == "linear":
                    taps = multirate_fir_linear(interpolation)
                elif taps == "linear-matlab":
                    taps = multirate_fir_linear_matlab(interpolation)
                elif taps == "zoh":
                    taps = multirate_fir_zoh(interpolation)

        if interpolation == 1:
            # PolyphaseFIR configured like Decimator
            super().__init__(decimation, taps, input="bottom-to-top", output="sum", streaming=streaming)
        else:
            # PolyphaseFIR configured like Interpolator
            super().__init__(interpolation, taps, input="hold", output="top-to-bottom", streaming=streaming)
            self._decimation *= decimation  # Due to downsampling output of Interpolator

    ##############################################################################
    # Special methods
    ##############################################################################

    def __call__(self, x: npt.ArrayLike, mode: Literal["rate", "full"] = "rate") -> npt.NDArray:
        y = super().__call__(x, mode)
        if self.interpolation > 1:
            y = y[:: self.decimation]  # Downsample the interpolated output
        return y

    def __repr__(self) -> str:
        if self.method == "custom":
            h_str = np.array2string(self.taps, max_line_width=int(1e6), separator=", ", suppress_small=True)
        else:
            h_str = repr(self.method)
        return (
            f"sdr.{type(self).__name__}({self.interpolation}, {self.decimation}, {h_str}, streaming={self.streaming})"
        )

    def __str__(self) -> str:
        string = f"sdr.{type(self).__name__}:"
        string += f"\n  order: {self.order}"
        string += f"\n  rate: {self.interpolation} / {self.decimation}"
        string += f"\n  method: {self.method!r}"
        string += f"\n  polyphase_taps: {self.polyphase_taps.shape} shape"
        string += f"\n  delay: {self.delay}"
        string += f"\n  streaming: {self.streaming}"
        return string

    ##############################################################################
    # Properties
    ##############################################################################

    @property
    def method(self) -> Literal["kaiser", "linear", "linear-matlab", "zoh", "custom"]:
        """
        The method used to design the polyphase resampling filter.
        """
        return self._method


@export
class Channelizer(PolyphaseFIR):
    r"""
    Implements a polyphase channelizer FIR filter.

    Notes:
        The polyphase channelizer efficiently splits the input signal $x[n]$ with sample rate $f_s$ into $C$
        equally spaced channels. Each channel has a bandwidth of $f_s / C$.

        The polyphase channelizer is equivalent to first mixing the input signal $x[n]$ with $C$ complex exponentials
        with frequencies $f_i = -i \cdot f_s / C$, filtering the mixed signals with the prototype FIR filter
        with feedforward coefficients $h[n]$, and then downsampling the filtered signals by $C$ (by discarding $C-1$
        samples every $C$ samples).

        Instead, the polyphase channelizer first decomposes the prototype FIR filter into $C$ polyphase filters with
        feedforward coefficients $h_i[n]$. The polyphase filters are then applied to the commutated input signal $x[n]$
        in parallel. The outputs of the polyphase filters are then inverse Discrete Fourier transformed (IDFT) to
        produce the $C$ channelized output signals $y_i[n]$.

        .. code-block:: text
           :caption: Polyphase Channelizer FIR Filter Block Diagram

                                                                  +------+
                                     +------------------------+   |      |
            ..., x[6], x[3], x[0] -->| h[0], h[3], h[6], h[9] |-->|      |--> ..., y[0,1], y[0,0]
                                     +------------------------+   |      |
                                     +------------------------+   |      |
            ..., x[5], x[2], 0    -->| h[1], h[4], h[7], 0    |-->| IDFT |--> ..., y[1,1], y[1,0]
                                     +------------------------+   |      |
                                     +------------------------+   |      |
            ..., x[4], x[1], 0    -->| h[2], h[5], h[8], 0    |-->|      |--> ..., y[2,1], y[2,0]
                                     +------------------------+   |      |
                                                                  +------+

            Input Commutator                                                 Parallel Outputs
            (bottom-to-top)

            x[n] = Input signal with sample rate fs
            y[i,n] = Channel i output signal with sample rate fs / C
            h[n] = Prototype FIR filter

        The polyphase feedforward taps $h_i[n]$ are related to the prototype feedforward taps $h[n]$ by

        $$h_i[j] = h[i + j C] .$$

    References:
        - fred harris, *Multirate Signal Processing for Communication Systems*, Chapter 6.1: Channelizer.

    Examples:
        Create a channelizer with 10 channels.

        .. ipython:: python

            C = 10
            channelizer = sdr.Channelizer(C); channelizer

        Create an input signal. Each channel has a tone with increasing frequency. The amplitude of each tone also
        increases by 2 dB for each channel.

        .. ipython:: python

            x = np.random.randn(10_000) + 1j * np.random.randn(10_000)
            for i in range(C):
                x += sdr.linear(10 + 2 * i) * np.exp(1j * 2 * np.pi * (i + 0.25 / C * i) / C * np.arange(10_000))

        Plot the input signal and overlay the channel boundaries. Note, Channel 5 is centered at $f = 0.5$. So, it
        wraps from positive to negative frequencies.

        .. ipython:: python

            plt.figure(); \
            sdr.plot.periodogram(x, fft=1024, color="k", label="Input $x[n]$");
            for i in range(C):
                f_start = (i - 0.5) / C
                f_stop = (i + 0.5) / C
                if f_start > 0.5:
                    f_start -= 1
                    f_stop -= 1
                plt.fill_betweenx([0, 80], f_start, f_stop, alpha=0.2, label=f"Channel {i}")
            @savefig sdr_Channelizer_1.png
            plt.xticks(np.arange(-0.5, 0.6, 0.1)); \
            plt.legend(); \
            plt.title("Input signals spread across 10 channels");

        Channelize the input signal with sample rate $f_s$ into 10 channels, each with sample rate $f_s / 10$.

        .. ipython:: python

            Y = channelizer(x)
            x.shape, Y.shape

        .. ipython:: python

            plt.figure();
            for i in range(C):
                sdr.plot.periodogram(Y[i, :], fft=1024, label=f"Channel {i}")
            @savefig sdr_Channelizer_2.png
            plt.xticks(np.arange(-0.5, 0.6, 0.1)); \
            plt.title("Output signals from 10 channels");

    Group:
        dsp-polyphase-filtering
    """

    def __init__(
        self,
        channels: int,
        taps: Literal["kaiser"] | npt.ArrayLike = "kaiser",
        polyphase_order: int = 23,
        atten: float = 80,
        streaming: bool = False,
    ):
        r"""
        Creates a polyphase FIR channelizing filter.

        Arguments:
            channels: The number of channels $C$.
            taps: The prototype filter design specification.

                - `"kaiser"`: The prototype filter is designed using :func:`sdr.multirate_fir`
                  with arguments 1 and `rate`.
                - `npt.ArrayLike`: The prototype filter feedforward coefficients $h[n]$.

            polyphase_order: The order of each polyphase filter. Must be odd, such that the filter lengths are even.
                Only used when `taps="kaiser"`.
            atten: The stopband attenuation $A_{\text{stop}}$ in dB. Only used when `taps="kaiser"`.
            streaming: Indicates whether to use streaming mode. In streaming mode, previous inputs are
                preserved between calls to :meth:`~Channelizer.__call__()`.
        """
        self._channels = verify_scalar(channels, int=True, positive=True)

        self._method: Literal["kaiser", "custom"]
        if not isinstance(taps, str):
            self._method = "custom"
            taps = verify_arraylike(taps, atleast_1d=True, ndim=1)
        else:
            self._method = verify_literal(taps, ["kaiser"])
            taps = multirate_fir(1, channels, polyphase_order, atten)

        super().__init__(channels, taps, input="bottom-to-top", output="all", streaming=streaming)

    ##############################################################################
    # Special methods
    ##############################################################################

    def __call__(
        self,
        x: npt.ArrayLike,
        mode: Literal["rate", "full"] = "rate",
    ) -> npt.NDArray:
        r"""
        Channelizes the input signal $x[n]$ with the polyphase FIR filter.

        Arguments:
            x: The input signal $x[n]$ with sample rate $f_s$ and length $L$.
            mode: The non-streaming convolution mode.

                - `"rate"`: The output signals $y_i[n]$ have length $L / C$, proportional to the number of channels $C$.
                  Output sample 0 aligns with input sample 0.
                - `"full"`: The full convolution is performed. The output signals $y_i[n]$ have length $(L + N) / C$,
                  where $N$ is the order of the multirate filter. Output sample :obj:`~sdr.Channelizer.delay` aligns
                  with input sample 0.

                In streaming mode, the `"full"` convolution is performed. However, for each $L$ input samples
                only $L / C$ output samples are produced per call. A final call to :meth:`~Channelizer.flush()`
                is required to flush the filter state.

        Returns:
            A 2-D array of channelized signals $y_i[n]$ with sample rate $f_s / C$. The output length is dictated by
            the `mode` argument.
        """
        Y = super().__call__(x, mode)
        Y = np.fft.ifft(Y, axis=0)
        return Y

    def __repr__(self) -> str:
        if self.method == "custom":
            h_str = np.array2string(self.taps, max_line_width=int(1e6), separator=", ", suppress_small=True)
        else:
            h_str = repr(self.method)
        return f"sdr.{type(self).__name__}({self.channels}, {h_str}, streaming={self.streaming})"

    def __str__(self) -> str:
        string = f"sdr.{type(self).__name__}:"
        string += f"\n  order: {self.order}"
        string += f"\n  channels: {self.channels}"
        string += f"\n  rate: {self.interpolation} / {self.decimation}"
        string += f"\n  method: {self.method!r}"
        string += f"\n  polyphase_taps: {self.polyphase_taps.shape} shape"
        string += f"\n  delay: {self.delay}"
        string += f"\n  streaming: {self.streaming}"
        return string

    ##############################################################################
    # Properties
    ##############################################################################

    @property
    def channels(self) -> float:
        """
        The number of channels $C$.
        """
        return self._channels

    @property
    def method(self) -> Literal["kaiser", "custom"]:
        """
        The method used to design the polyphase channelizing filter.
        """
        return self._method



"""
A module containing functions to estimate frequency-domain parameters.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from .._conversion import linear
from .._helper import convert_output, export, verify_arraylike
from ._snr import composite_snr


@export
def foa_crlb(
    snr: npt.ArrayLike,
    time: npt.ArrayLike,
    bandwidth: npt.ArrayLike,
    rms_integration_time: npt.ArrayLike | None = None,
    noise_bandwidth: npt.ArrayLike | None = None,
) -> npt.NDArray[np.float64]:
    r"""
    Calculates the Cramér-Rao lower bound (CRLB) on frequency of arrival (FOA) estimation.

    Arguments:
        snr: The signal-to-noise ratio (SNR) of the signal $\gamma = S / (N_0 B_n)$ in dB.
        time: The integration time $T$ in seconds.
        bandwidth: The signal bandwidth $B_s$ in Hz.
        rms_integration_time: The root-mean-square (RMS) integration time $T_{\text{rms}}$ in Hz. If `None`, the RMS
            integration time is calculated assuming a rectangular power envelope, $T_{\text{rms}} = T/\sqrt{12}$.
        noise_bandwidth: The noise bandwidth $B_n$ in Hz. If `None`, the noise bandwidth is assumed to be the
            signal bandwidth $B_s$.

    Returns:
        The Cramér-Rao lower bound (CRLB) on the frequency of arrival (FOA) estimation error standard
        deviation $\sigma_{\text{foa}}$ in Hz.

    See Also:
        sdr.rms_integration_time

    Notes:
        The Cramér-Rao lower bound (CRLB) on the frequency of arrival (FOA) estimation error standard
        deviation $\sigma_{\text{foa}}$ is given by

        $$\sigma_{\text{foa}} = \frac{1}{\pi \sqrt{8} T_{\text{rms}}} \frac{1}{\sqrt{B_n T \gamma}}$$

        $$
        T_{\text{rms}} = \sqrt{\frac
        {\int_{-\infty}^{\infty} (t - \mu_t)^2 \cdot \left| x(t - \mu_t) \right|^2 \, dt}
        {\int_{-\infty}^{\infty} \left| x(t - \mu_t) \right|^2 \, dt}
        }
        $$

        where $\gamma$ is the signal-to-noise ratio (SNR), $\left| x(t) \right|^2$ is the power envelope of the signal,
        and $\mu_t$ is the centroid of the power envelope.

        .. note::
            The constant terms from Stein's original equations were rearranged. The factor of 2 was removed from
            $\gamma$ and the factor of $2\pi$ was removed from $T_{\text{rms}}$ and incorporated into the CRLB
            equation.

        The signal-to-noise ratio (SNR) $\gamma$ is improved by the coherent integration gain, which is the
        time-bandwidth product $B_n T$. The product $B_n T \gamma$ is the output SNR of the matched filter
        or correlator, which is equivalent to $E / N_0$.

        $$B_n T \gamma = B_n T \frac{S}{N_0 B_n} = \frac{S T}{N_0} = \frac{E}{N_0}$$

        .. warning::
            According to Stein, the CRLB equation only holds for output SNRs greater than 10 dB. This ensures there is
            sufficient SNR to correctly identify the time/frequency peak without high $P_{fa}$. Given the rearrangement
            of scaling factors, CRLB values with output SNRs less than 7 dB are set to NaN.

        The frequency measurement precision is inversely proportional to the integration time of the signal and the
        square root of the output SNR.

    Examples:
        .. ipython:: python

            snr = 10
            bandwidth = np.logspace(5, 8, 101)

            @savefig sdr_foa_crlb_1.png
            plt.figure(); \
            plt.loglog(bandwidth, sdr.foa_crlb(snr, 1e-6, bandwidth), label="1 μs"); \
            plt.loglog(bandwidth, sdr.foa_crlb(snr, 1e-5, bandwidth), label="10 μs"); \
            plt.loglog(bandwidth, sdr.foa_crlb(snr, 1e-4, bandwidth), label="100 μs"); \
            plt.loglog(bandwidth, sdr.foa_crlb(snr, 1e-3, bandwidth), label="1 ms"); \
            plt.legend(title="Integration time"); \
            plt.xlim(1e5, 1e8); \
            plt.ylim(1e0, 1e6); \
            plt.xlabel("Bandwidth (Hz), $B$"); \
            plt.ylabel(r"CRLB on FOA (Hz), $\sigma_{\text{foa}}$"); \
            plt.title(f"Cramér-Rao lower bound (CRLB) on FOA estimation error\nstandard deviation with {snr}-dB SNR");

    Group:
        estimation-frequency
    """
    # The second signal's SNR of 1 million dB is equivalent to a noiseless template
    return fdoa_crlb(snr, 1_000_000, time, bandwidth, rms_integration_time, noise_bandwidth)


@export
def fdoa_crlb(
    snr1: npt.ArrayLike,
    snr2: npt.ArrayLike,
    time: npt.ArrayLike,
    bandwidth: npt.ArrayLike,
    rms_integration_time: npt.ArrayLike | None = None,
    noise_bandwidth: npt.ArrayLike | None = None,
) -> npt.NDArray[np.float64]:
    r"""
    Calculates the Cramér-Rao lower bound (CRLB) on frequency difference of arrival (FDOA) estimation.

    Arguments:
        snr1: The signal-to-noise ratio (SNR) of the first signal $\gamma_1 = S_1 / (N_0 B_n)$ in dB.
        snr2: The signal-to-noise ratio (SNR) of the second signal $\gamma_2 = S_2 / (N_0 B_n)$ in dB.
        time: The integration time $T$ in seconds.
        bandwidth: The signal bandwidth $B_s$ in Hz.
        rms_integration_time: The root-mean-square (RMS) integration time $T_{\text{rms}}$ in Hz. If `None`, the RMS
            integration time is calculated assuming a rectangular power envelope, $T_{\text{rms}} = T/\sqrt{12}$.
        noise_bandwidth: The noise bandwidth $B_n$ in Hz. If `None`, the noise bandwidth is assumed to be the
            signal bandwidth $B_s$. The noise bandwidth must be the same for both signals.

    Returns:
        The Cramér-Rao lower bound (CRLB) on the frequency difference of arrival (FDOA) estimation error standard
        deviation $\sigma_{\text{fdoa}}$ in Hz.

    See Also:
        sdr.rms_integration_time

    Notes:
        The Cramér-Rao lower bound (CRLB) on the frequency difference of arrival (FDOA) estimation error standard
        deviation $\sigma_{\text{fdoa}}$ is given by

        $$\sigma_{\text{fdoa}} = \frac{1}{\pi \sqrt{8} T_{\text{rms}}} \frac{1}{\sqrt{B_n T \gamma}}$$

        $$\frac{1}{\gamma} = \frac{1}{\gamma_1} + \frac{1}{\gamma_2} + \frac{1}{\gamma_1 \gamma_2}$$

        $$
        T_{\text{rms}} = \sqrt{\frac
        {\int_{-\infty}^{\infty} (t - \mu_t)^2 \cdot \left| x(t - \mu_t) \right|^2 \, dt}
        {\int_{-\infty}^{\infty} \left| x(t - \mu_t) \right|^2 \, dt}
        }
        $$

        where $\gamma$ is the effective signal-to-noise ratio (SNR), $\left| x(t) \right|^2$ is the power envelope of
        the signal, and $\mu_t$ is the centroid of the power envelope.

        .. note::
            The constant terms from Stein's original equations were rearranged. The factor of 2 was removed from
            $\gamma$ and the factor of $2\pi$ was removed from $T_{\text{rms}}$ and incorporated into the CRLB
            equation.

        The effective signal-to-noise ratio (SNR) $\gamma$ is improved by the coherent integration gain, which is the
        time-bandwidth product $B_n T$. The product $B_n T \gamma$ is the output SNR of the matched filter
        or correlator, which is equivalent to $E / N_0$.

        $$B_n T \gamma = B_n T \frac{S}{N_0 B_n} = \frac{S T}{N_0} = \frac{E}{N_0}$$

        .. warning::
            According to Stein, the CRLB equation only holds for output SNRs greater than 10 dB. This ensures there is
            sufficient SNR to correctly identify the time/frequency peak without high $P_{fa}$. Given the rearrangement
            of scaling factors, CRLB values with output SNRs less than 7 dB are set to NaN.

        The frequency measurement precision is inversely proportional to the integration time of the signal and the
        square root of the output SNR.

    Examples:
        .. ipython:: python

            snr = 10
            bandwidth = np.logspace(5, 8, 101)

            @savefig sdr_fdoa_crlb_1.png
            plt.figure(); \
            plt.loglog(bandwidth, sdr.fdoa_crlb(snr, snr, 1e-6, bandwidth), label="1 μs"); \
            plt.loglog(bandwidth, sdr.fdoa_crlb(snr, snr, 1e-5, bandwidth), label="10 μs"); \
            plt.loglog(bandwidth, sdr.fdoa_crlb(snr, snr, 1e-4, bandwidth), label="100 μs"); \
            plt.loglog(bandwidth, sdr.fdoa_crlb(snr, snr, 1e-3, bandwidth), label="1 ms"); \
            plt.legend(title="Integration time"); \
            plt.xlim(1e5, 1e8); \
            plt.ylim(1e0, 1e6); \
            plt.xlabel("Bandwidth (Hz), $B$"); \
            plt.ylabel(r"CRLB on FDOA (Hz), $\sigma_{\text{fdoa}}$"); \
            plt.title(f"Cramér-Rao lower bound (CRLB) on FDOA estimation error\nstandard deviation with {snr}-dB SNR");

    Group:
        estimation-frequency
    """
    snr1 = verify_arraylike(snr1, float=True)
    snr2 = verify_arraylike(snr2, float=True)
    time = verify_arraylike(time, float=True, positive=True)
    bandwidth = verify_arraylike(bandwidth, float=True, positive=True)
    if rms_integration_time is None:
        rms_integration_time = time / np.sqrt(12)
    else:
        rms_integration_time = verify_arraylike(rms_integration_time, float=True, positive=True)
    if noise_bandwidth is None:
        noise_bandwidth = bandwidth
    else:
        noise_bandwidth = verify_arraylike(noise_bandwidth, float=True, positive=True)

    # The effective SNR is improved by the coherent integration gain, which is the time-bandwidth product
    snr = composite_snr(snr1, snr2)
    snr = linear(snr)
    output_snr = time * noise_bandwidth * snr

    # Stein specifically mentions that the equations are only valid for output SNR greater than 10 dB.
    # Since we factored 2 out from the composite SNR, we need to compare against 7 dB.
    output_snr = np.where(output_snr >= linear(7), output_snr, np.nan)

    crlb = 1 / (np.pi * np.sqrt(8) * rms_integration_time * np.sqrt(output_snr))

    return convert_output(crlb)



import numpy as np
import pytest

import sdr


def test_peak_voltage():
    rng = np.random.default_rng()
    A = rng.uniform(2, 10)  # Signal amplitude
    N = 50
    x = A * np.sin(np.linspace(0, 2 * np.pi, N, endpoint=False))
    assert sdr.peak_voltage(x) == pytest.approx(A, rel=1e-2)


def test_rms_voltage():
    rng = np.random.default_rng()
    A = rng.uniform(2, 10)  # Signal amplitude
    N = 50
    x = A * np.sin(np.linspace(0, 2 * np.pi, N, endpoint=False))
    assert sdr.rms_voltage(x) == pytest.approx(A / np.sqrt(2), rel=1e-2)



import numpy as np
import pytest

import sdr


def test_real_from_snr():
    rng = np.random.default_rng()
    A = rng.uniform(2, 3)  # Signal amplitude
    N = 1000
    x = A * np.ones(N, dtype=float)  # Use a constant signal so the power measurement has no uncertainty

    snr = rng.uniform(1, 2)  # dB
    snr_linear = 10 ** (snr / 10)
    signal_power = A**2  # Theoretical power of a constant signal
    noise_std = np.sqrt(signal_power / snr_linear)

    y = sdr.awgn(x, snr=snr)

    # Use standard deviation so the measurement errors aren't compounded
    std = np.std(y - x)
    assert std == pytest.approx(noise_std, rel=1e-1)


def test_complex_from_snr():
    rng = np.random.default_rng()
    A = rng.uniform(2, 3)  # Signal amplitude
    N = 1000
    x = A * np.ones(N, dtype=complex)  # Use a constant signal so the power measurement has no uncertainty

    snr = rng.uniform(1, 2)  # dB
    snr_linear = 10 ** (snr / 10)
    signal_power = A**2  # Theoretical power of a constant signal
    noise_std = np.sqrt(signal_power / snr_linear)

    y = sdr.awgn(x, snr=snr)

    # Use standard deviation so the measurement errors aren't compounded
    std = np.sqrt(np.var((y - x).real) + np.var((y - x).imag))
    assert std == pytest.approx(noise_std, rel=1e-1)


def test_real_from_noise():
    rng = np.random.default_rng()
    A = rng.uniform(2, 3)  # Signal amplitude
    N = 1000
    x = A * np.ones(N, dtype=float)  # Use a constant signal so the power measurement has no uncertainty

    noise_std = rng.uniform(1, 2)
    y = sdr.awgn(x, noise=noise_std**2)

    # Use standard deviation so the measurement errors aren't compounded
    std = np.std(y - x)
    assert std == pytest.approx(noise_std, rel=1e-1)


def test_complex_from_noise():
    rng = np.random.default_rng()
    A = rng.uniform(2, 3)  # Signal amplitude
    N = 1000
    x = A * np.ones(N, dtype=complex)  # Use a constant signal so the power measurement has no uncertainty

    noise_std = rng.uniform(1, 2)
    y = sdr.awgn(x, noise=noise_std**2)

    # Use standard deviation so the measurement errors aren't compounded
    std = np.sqrt(np.var((y - x).real) + np.var((y - x).imag))
    assert std == pytest.approx(noise_std, rel=1e-1)


def test_seed():
    x = np.ones(100)
    y1 = sdr.awgn(x, snr=10, seed=0)
    y2 = sdr.awgn(x, snr=10, seed=0)
    assert np.all(y1 == y2)



"""
A module containing classes for minimum-shift keying (MSK) modulations.
"""

from __future__ import annotations

import numpy.typing as npt
from typing_extensions import Literal

from .._helper import export, verify_scalar
from ._psk import OQPSK
from ._pulse_shapes import half_sine


@export
class MSK(OQPSK):
    r"""
    Implements minimum-shift keying (MSK) modulation and demodulation.

    Notes:
        MSK is a linear phase modulation scheme similar to OQPSK. One key distinction is that the pulse
        shape is a half sine wave. This results in a constant envelope signal, which results in a lower
        peak-to-average power ratio (PAPR).

        MSK can also be consider as continuous-phase frequency-shift keying (CPFSK) with the frequency separation
        equaling half the bit period.

    .. nomenclature::
        :collapsible:

        - $k$: Symbol index
        - $n$: Sample index
        - $s[k]$: Decimal symbols
        - $a[k]$ Complex symbols
        - $x[n]$: Pulse-shaped complex samples
        - $\tilde{x}[n]$: Received (noisy) pulse-shaped complex samples
        - $\tilde{a}[k]$: Received (noisy) complex symbols
        - $\hat{a}[k]$: Complex symbol decisions
        - $\hat{s}[k]$: Decimal symbol decisions

    Examples:
        Create a MSK modem.

        .. ipython:: python

            msk = sdr.MSK(); msk

            @savefig sdr_MSK_1.png
            plt.figure(); \
            sdr.plot.symbol_map(msk);

        Generate a random bit stream, convert to 2-bit symbols, and map to complex symbols.

        .. ipython:: python

            bits = np.random.randint(0, 2, 1000); bits[0:8]
            symbols = sdr.pack(bits, msk.bps); symbols[0:4]
            complex_symbols = msk.map_symbols(symbols); complex_symbols[0:4]

            @savefig sdr_MSK_2.png
            plt.figure(); \
            sdr.plot.constellation(complex_symbols, linestyle="-");

        Modulate and pulse shape the symbols to a complex baseband signal.

        .. ipython:: python

            tx_samples = msk.modulate(symbols)

            @savefig sdr_MSK_3.png
            plt.figure(); \
            sdr.plot.time_domain(tx_samples[0:50*msk.sps]);

        MSK, like OQPSK, has I and Q channels that are offset by half a symbol period.

        .. ipython:: python

            @savefig sdr_MSK_4.png
            plt.figure(figsize=(8, 6)); \
            sdr.plot.eye(tx_samples[5*msk.sps : -5*msk.sps], msk.sps); \
            plt.suptitle("Noiseless transmitted signal");

        The phase trajectory of MSK is linear and continuous. Although, it should be noted that the phase is not
        differentiable at the symbol boundaries. This leads to lower spectral efficiency than, for instance,
        GMSK.

        .. ipython:: python

            @savefig sdr_MSK_5.png
            plt.figure(); \
            sdr.plot.phase_tree(tx_samples[msk.sps:], msk.sps);

        Add AWGN noise such that $E_b/N_0 = 30$ dB.

        .. ipython:: python

            ebn0 = 30; \
            snr = sdr.ebn0_to_snr(ebn0, bps=msk.bps, sps=msk.sps); \
            rx_samples = sdr.awgn(tx_samples, snr=snr)

            @savefig sdr_MSK_6.png
            plt.figure(); \
            sdr.plot.time_domain(rx_samples[0:50*msk.sps]);

        Manually apply a matched filter. Examine the eye diagram of the matched filtered received signal.

        .. ipython:: python

            mf = sdr.FIR(msk.pulse_shape); \
            mf_samples = mf(rx_samples)

            @savefig sdr_MSK_7.png
            plt.figure(figsize=(8, 6)); \
            sdr.plot.eye(mf_samples[10*msk.sps : -10*msk.sps], msk.sps); \
            plt.suptitle("Noisy received and matched filtered signal");

        Matched filter and demodulate. Note, the first symbol has $Q = 0$ and the last symbol has $I = 0$.

        .. ipython:: python

            rx_symbols, rx_complex_symbols, _ = msk.demodulate(rx_samples)

            # The symbol decisions are error-free
            np.array_equal(symbols, rx_symbols)

            @savefig sdr_MSK_8.png
            plt.figure(); \
            sdr.plot.constellation(rx_complex_symbols);

        See the :ref:`psk` example.

    Group:
        modulation-continuous-phase
    """

    def __init__(
        self,
        phase_offset: float = 45,
        symbol_labels: Literal["bin", "gray"] | npt.ArrayLike = "gray",
        sps: int = 8,
    ):
        r"""
        Creates a new MSK object.

        Arguments:
            phase_offset: The absolute phase offset $\phi$ in degrees.
            symbol_labels: The decimal symbol labels of consecutive complex symbols.

                - `"bin"`: The symbols are binary-coded. Adjacent symbols may differ by more than one bit.
                - `"gray":` The symbols are Gray-coded. Adjacent symbols only differ by one bit.
                - `npt.ArrayLike`: An $4$-length array whose indices are the default symbol labels and whose values are
                  the new symbol labels. The default symbol labels are $0$ to $4-1$ for phases starting at $1 + 0j$
                  and going counter-clockwise around the unit circle.

            sps: The number of samples per symbol $f_s / f_{sym}$.

        See Also:
            sdr.half_sine
        """
        verify_scalar(phase_offset, float=True)
        verify_scalar(sps, int=True, positive=True, even=True)

        pulse_shape = half_sine(sps)

        super().__init__(
            phase_offset=phase_offset,
            symbol_labels=symbol_labels,
            sps=sps,
            pulse_shape=pulse_shape,
        )



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html


# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import inspect
import os
import re
import sys

sys.path.insert(0, os.path.abspath("."))
sys.path.insert(0, os.path.abspath(".."))

# Need to build docs with Python 3.8 or higher for proper typing annotations, including from __future__ import annotations
assert sys.version_info.major == 3 and sys.version_info.minor >= 8

# Assign a build variable to the builtin module that inerts the @set_module decorator. This is done because set_module
# confuses Sphinx when parsing overloaded functions. When not building the documentation, the @set_module("sdr")
# decorator works as intended.
import builtins

setattr(builtins, "__sphinx_build__", True)

import numpy
import sphinx

import sdr

# -- Project information -----------------------------------------------------

project = "sdr"
copyright = "2023-2024, Matt Hostetter"
author = "Matt Hostetter"
version = sdr.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named "sphinx.ext.*") or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.mathjax",
    "sphinx.ext.intersphinx",
    "sphinx_last_updated_by_git",
    "sphinx_immaterial",
    "sphinx_immaterial.apidoc.python.apigen",
    "sphinx_math_dollar",
    # "myst_parser",
    "myst_nb",
    "sphinx_design",
    "IPython.sphinxext.ipython_console_highlighting",
    "IPython.sphinxext.ipython_directive",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = [".rst", ".md", ".ipynb"]

# Tell sphinx that ReadTheDocs will create an index.rst file as the main file,
# not the default of contents.rst.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", ".ipynb_checkpoints"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_immaterial"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_css_files = [
    "extra.css",
]

# Define a custom inline Python syntax highlighting literal
rst_prolog = """
.. role:: python(code)
   :language: python
   :class: highlight
"""

# Sets the default role of `content` to :python:`content`, which uses the custom Python syntax highlighting inline literal
default_role = "python"

html_title = "sdr"
# html_favicon = "../logo/sdr-favicon-color.png"
# html_logo = "../logo/sdr-favicon-white.png"

# Sphinx Immaterial theme options
html_theme_options = {
    "icon": {
        "repo": "fontawesome/brands/github",
    },
    "site_url": "https://mhostetter.github.io/sdr/latest/",
    "repo_url": "https://github.com/mhostetter/sdr",
    "repo_name": "mhostetter/sdr",
    "social": [
        {
            "icon": "fontawesome/brands/github",
            "link": "https://github.com/mhostetter/sdr",
        },
        {
            "icon": "fontawesome/brands/python",
            "link": "https://pypi.org/project/sdr/",
        },
        {
            "icon": "fontawesome/brands/twitter",
            "link": "https://twitter.com/sdr_py",
        },
    ],
    "edit_uri": "",
    "globaltoc_collapse": False,
    "features": [
        # "navigation.expand",
        "navigation.tabs",
        # "toc.integrate",
        "navigation.sections",
        # "navigation.instant",
        # "header.autohide",
        "navigation.top",
        "navigation.tracking",
        "toc.follow",
        "toc.sticky",
        "content.tabs.link",
        "announce.dismiss",
    ],
    "palette": [
        {
            "media": "(prefers-color-scheme: light)",
            "scheme": "default",
            "primary": "indigo",
            "accent": "indigo",
            "toggle": {
                "icon": "material/weather-night",
                "name": "Switch to dark mode",
            },
        },
        {
            "media": "(prefers-color-scheme: dark)",
            "scheme": "slate",
            "primary": "black",
            "accent": "indigo",
            "toggle": {
                "icon": "material/weather-sunny",
                "name": "Switch to light mode",
            },
        },
    ],
    "analytics": {
        "provider": "google",
        "property": "G-PVMFVF4016",
    },
    "version_dropdown": True,
    "version_json": "../versions.json",
}

html_last_updated_fmt = ""
html_use_index = True
html_domain_indices = True

# TODO: Disable nitpick for now, re-enable later
nitpicky = False
nitpick_ignore = [
    # Wish these would work
    ("py:class", "numpy.typing.ArrayLike"),
    ("py:class", "numpy.typing.DTypeLike"),
    ("py:class", "numpy.typing.NDArray"),
    ("py:class", "numpy.int64"),
    ("py:class", "numpy.float64"),
    ("py:class", "numpy.complex128"),
]


# -- Extension configuration -------------------------------------------------

# Create hyperlinks to other documentation
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "galois": ("https://mhostetter.github.io/galois/latest/", None),
}

autodoc_default_options = {
    "imported-members": True,
    "members": True,
    # "special-members": True,
    # "inherited-members": "ndarray",
    # "member-order": "groupwise",
}
autodoc_typehints = "signature"
autodoc_typehints_description_target = "documented"
autodoc_typehints_format = "short"

autodoc_type_aliases = {
    "npt.ArrayLike": "~numpy.typing.ArrayLike",
    "npt.DTypeLike": "~numpy.typing.DTypeLike",
    "npt.NDArray": "~numpy.typing.NDArray",
}

ipython_execlines = [
    "import math",
    "import matplotlib.pyplot as plt",
    "import numpy as np",
    "import scipy.stats",
    "import sdr",
    "import galois",
    "np.random.seed(0)",  # Set random seed for reproducibility
    "rng = np.random.default_rng(0)",  # Set random seed for reproducibility
    "sdr.plot.use_style()",  # Use the sdr library matplotlib style
]
ipython_warning_is_error = True

myst_enable_extensions = ["dollarmath"]

nb_execution_mode = "off"
# nb_output_stderr = "remove-warn"

mathjax_config = {
    "tex2jax": {
        "inlineMath": [["\\(", "\\)"]],
        "displayMath": [["\\[", "\\]"]],
    },
}
mathjax3_config = {
    "tex": {
        "inlineMath": [["\\(", "\\)"]],
        "displayMath": [["\\[", "\\]"]],
    }
}

# -- Sphinx Immaterial configs -------------------------------------------------

# Python apigen configuration
python_apigen_modules = {
    "sdr": "api/sdr.",
    "sdr.plot": "api/sdr.plot.",
}
python_apigen_default_groups = [
    ("class:.*", "Classes"),
    ("data:.*", "Variables"),
    ("function:.*", "Functions"),
    ("classmethod:.*", "Class methods"),
    ("method:.*", "Methods"),
    (r"method:.*\.[A-Z][A-Za-z,_]*", "Constructors"),
    (r"method:.*\.__[A-Za-z,_]*__", "Special methods"),
    (r"method:.*\.__(init|new)__", "Constructors"),
    (r"method:.*\.__(str|repr)__", "String representation"),
    (r"method:.*\.(plot).*", "Plotting"),
    ("property:.*", "Properties"),
    (r".*:.*\.is_[a-z,_]*", "Attributes"),
]
python_apigen_default_order = [
    ("class:.*", 10),
    ("data:.*", 11),
    ("function:.*", 12),
    ("classmethod:.*", 40),
    ("method:.*", 50),
    (r"method:.*\.[A-Z][A-Za-z,_]*", 20),
    (r"method:.*\.__[A-Za-z,_]*__", 28),
    (r"method:.*\.__(init|new)__", 20),
    (r"method:.*\.__(str|repr)__", 30),
    (r"method:.*\.(plot).*", 51),
    ("property:.*", 60),
    (r".*:.*\.is_[a-z,_]*", 70),
]
# python_apigen_order_tiebreaker = "alphabetical"
python_apigen_case_insensitive_filesystem = False
python_apigen_show_base_classes = True

# Python domain directive configuration
python_type_aliases = autodoc_type_aliases
python_module_names_to_strip_from_xrefs = ["collections.abc"]

# General API configuration
object_description_options = [
    ("py:.*", dict(include_rubrics_in_toc=True)),
]

sphinx_immaterial_custom_admonitions = [
    {
        "name": "note",
        "title": "Note",
        "classes": ["collapsible"],
        "icon": "fontawesome/solid/pencil",
        "override": True,
    },
    {
        "name": "warning",
        "title": "Warning",
        "classes": ["collapsible"],
        "icon": "fontawesome/solid/exclamation",
        "override": True,
    },
    {
        "name": "info",
        "icon": "fontawesome/solid/circle-info",
        "override": True,
    },
    {
        "name": "tip",
        "icon": "fontawesome/regular/lightbulb",
        "override": True,
    },
    {
        "name": "abstract",
        "icon": "fontawesome/regular/file-lines",
        "override": True,
    },
    {
        "name": "important",
        "icon": "fontawesome/solid/bolt",
        "override": True,
    },
    {
        "name": "example",
        "icon": "fontawesome/solid/terminal",
        "override": True,
    },
    {
        "name": "quote",
        "icon": "fontawesome/solid/quote-left",
        "override": True,
    },
    {
        "name": "question",
        "title": "Question",
        "classes": ["collapsible"],
        "icon": "fontawesome/solid/question",
        "color": (108, 117, 125),  # --sd-color-secondary
        "override": True,
    },
    {
        "name": "seealso",
        "title": "See also",
        "classes": ["collapsible"],
        "icon": "fontawesome/solid/magnifying-glass",
        "color": (108, 117, 125),  # --sd-color-secondary
        "override": True,
    },
    {
        "name": "versionadded",
        "icon": "fontawesome/solid/code-commit",
        "override": True,
    },
    {
        "name": "versionchanged",
        "icon": "fontawesome/solid/code-branch",
        "override": True,
    },
    {
        # This needs to be defined here so the icon is available when referenced in _templates/base.html
        "name": "star",
        "icon": "fontawesome/regular/star",
        "color": (255, 233, 3),  # Gold
    },
    {
        "name": "nomenclature",
        "title": "Variable nomenclature",
        "classes": ["collapsible"],
        "icon": "fontawesome/solid/arrow-down-a-z",
        "color": (108, 117, 125),  # --sd-color-secondary
    },
    {
        "name": "fast-performance",
        "title": "Faster performance",
        "icon": "material/speedometer",
        "color": (47, 177, 112),  # Green: --md-code-hl-string-color
    },
    {
        "name": "slow-performance",
        "title": "Slower performance",
        "icon": "material/speedometer-slow",
        "color": (230, 105, 91),  # Red: --md-code-hl-number-color
    },
]


# -- Monkey-patching ---------------------------------------------------------

# SPECIAL_MEMBERS = [
#     "__repr__",
#     "__str__",
#     "__int__",
#     "__call__",
#     "__len__",
#     "__eq__",
# ]


def autodoc_skip_member(app, what, name, obj, skip, options):
    """
    Instruct autodoc to skip members that are inherited from np.ndarray.
    """
    if name in ["__str__", "__repr__"]:
        return True

    return skip


def autodoc_process_bases(app, name, obj, options, bases):
    """
    Remove private classes or mixin classes from documented class bases.
    """
    # Determine the bases to be removed
    remove_bases = []
    for base in bases:
        if base.__name__[0] == "_" or "Mixin" in base.__name__:
            remove_bases.append(base)

    # Remove from the bases list in-place
    for base in remove_bases:
        bases.remove(base)


# # Only during Sphinx builds, monkey-patch the metaclass properties into this class as "class properties". In Python 3.9 and greater,
# # class properties may be created using `@classmethod @property def foo(cls): return "bar"`. In earlier versions, they must be created
# # in the metaclass, however Sphinx cannot find or document them. Adding this workaround allows Sphinx to document them.


# def classproperty(obj):
#     ret = classmethod(obj)
#     ret.__doc__ = obj.__doc__
#     return ret


# ArrayMeta_properties = [
#     member for member in dir(sdr.Array) if inspect.isdatadescriptor(getattr(type(sdr.Array), member, None))
# ]
# for p in ArrayMeta_properties:
#     # Fetch the class properties from the private metaclasses
#     ArrayMeta_property = getattr(sdr._domains._meta.ArrayMeta, p)

#     # Temporarily delete the class properties from the private metaclasses
#     delattr(sdr._domains._meta.ArrayMeta, p)

#     # Add a Python 3.9 style class property to the public class
#     setattr(sdr.Array, p, classproperty(ArrayMeta_property))

#     # Add back the class properties to the private metaclasses
#     setattr(sdr._domains._meta.ArrayMeta, p, ArrayMeta_property)


# FieldArrayMeta_properties = [
#     member for member in dir(sdr.FieldArray) if inspect.isdatadescriptor(getattr(type(sdr.FieldArray), member, None))
# ]
# for p in FieldArrayMeta_properties:
#     # Fetch the class properties from the private metaclasses
#     if p in ArrayMeta_properties:
#         ArrayMeta_property = getattr(sdr._domains._meta.ArrayMeta, p)
#     FieldArrayMeta_property = getattr(sdr._fields._meta.FieldArrayMeta, p)

#     # Temporarily delete the class properties from the private metaclasses
#     if p in ArrayMeta_properties:
#         delattr(sdr._domains._meta.ArrayMeta, p)
#     delattr(sdr._fields._meta.FieldArrayMeta, p)

#     # Add a Python 3.9 style class property to the public class
#     setattr(sdr.FieldArray, p, classproperty(FieldArrayMeta_property))

#     # Add back the class properties to the private metaclasses
#     if p in ArrayMeta_properties:
#         setattr(sdr._domains._meta.ArrayMeta, p, ArrayMeta_property)
#     setattr(sdr._fields._meta.FieldArrayMeta, p, FieldArrayMeta_property)


def autodoc_process_signature(app, what, name, obj, options, signature, return_annotation):
    signature = modify_type_hints(signature)
    return_annotation = modify_type_hints(return_annotation)
    return signature, return_annotation


def modify_type_hints(signature):
    """
    Fix shortening numpy type annotations in string annotations created with
    `from __future__ import annotations` that Sphinx can't process before Python
    3.10.

    See https://github.com/jbms/sphinx-immaterial/issues/161
    """
    if signature:
        signature = re.sub(r"(?<!~)np\.", "~numpy.", signature)
    return signature


def monkey_patch_parse_see_also():
    """
    Use the NumPy docstring parsing of See Also sections for convenience. This automatically
    hyperlinks plaintext functions and methods.
    """
    # Add the special parsing method from NumpyDocstring
    method = sphinx.ext.napoleon.NumpyDocstring._parse_numpydoc_see_also_section
    sphinx.ext.napoleon.GoogleDocstring._parse_numpydoc_see_also_section = method

    def _parse_see_also_section(self, section: str):
        """Copied from NumpyDocstring._parse_see_also_section()."""
        lines = self._consume_to_next_section()

        # Added: strip whitespace from lines to satisfy _parse_numpydoc_see_also_section()
        for i in range(len(lines)):
            lines[i] = lines[i].strip()

        try:
            return self._parse_numpydoc_see_also_section(lines)
        except ValueError:
            return self._format_admonition("seealso", lines)

    sphinx.ext.napoleon.GoogleDocstring._parse_see_also_section = _parse_see_also_section


def setup(app):
    monkey_patch_parse_see_also()

    app.connect("autodoc-skip-member", autodoc_skip_member)
    app.connect("autodoc-process-bases", autodoc_process_bases)
    app.connect("autodoc-process-signature", autodoc_process_signature)



"""
A module containing a base class for linear phase/amplitude modulations.
"""

from __future__ import annotations

import abc

import numpy as np
import numpy.typing as npt
from typing_extensions import Literal

from .._filter import Decimator, Interpolator
from .._helper import convert_output, export, verify_arraylike, verify_literal, verify_scalar
from ._pulse_shapes import raised_cosine, rectangular, root_raised_cosine


@export
class LinearModulation:
    r"""
    Implements linear phase/amplitude modulation with arbitrary symbol mapping.

    .. nomenclature::
        :collapsible:

        - $k$: Symbol index
        - $n$: Sample index
        - $s[k]$: Decimal symbols
        - $a[k]$ Complex symbols
        - $x[n]$: Pulse-shaped complex samples
        - $\tilde{x}[n]$: Received (noisy) pulse-shaped complex samples
        - $\tilde{a}[k]$: Received (noisy) complex symbols
        - $\hat{a}[k]$: Complex symbol decisions
        - $\hat{s}[k]$: Decimal symbol decisions

    Group:
        modulation-linear
    """

    def __init__(
        self,
        symbol_map: npt.ArrayLike,
        phase_offset: float = 0.0,
        sps: int = 8,
        pulse_shape: npt.ArrayLike | Literal["rect", "rc", "srrc"] = "rect",
        span: int | None = None,
        alpha: float | None = None,
    ):
        r"""
        Creates a new linear phase/amplitude modulation object.

        Arguments:
            symbol_map: The symbol mapping $\{0, \dots, M-1\} \mapsto \mathbb{C}$. An $M$-length array whose indices
                are decimal symbols $s[k]$ and whose values are complex symbols $a[k]$, where $M$ is the
                modulation order.
            phase_offset: A phase offset $\phi$ in degrees to apply to `symbol_map`.
            sps: The number of samples per symbol $f_s / f_{sym}$.
            pulse_shape: The pulse shape $h[n]$ of the modulated signal.

                - `npt.ArrayLike`: A custom pulse shape. It is important that `sps` matches the design
                  of the pulse shape. See :ref:`pulse-shaping-functions`.
                - `"rect"`: Rectangular pulse shape.
                - `"rc"`: Raised cosine pulse shape.
                - `"srrc"`: Square-root raised cosine pulse shape.

            span: The span of the pulse shape in symbols. This is only used if `pulse_shape` is a string.
                If `None`, 1 is used for `"rect"` and 10 is used for `"rc"` and `"srrc"`.
            alpha: The roll-off factor of the pulse shape. If `None`, 0.2 is used for `"rc"` and `"srrc"`.

        See Also:
            sdr.rectangular, sdr.raised_cosine, sdr.root_raised_cosine
        """
        self._symbol_map = verify_arraylike(symbol_map, complex=True, ndim=1)  # Decimal-to-complex symbol map
        self._order = verify_scalar(self._symbol_map.size, power_of_two=True)  # Modulation order
        self._bps = int(np.log2(self._order))  # Coded bits per symbol
        self._phase_offset = verify_scalar(phase_offset, float=True)  # Phase offset in degrees
        self._sps = verify_scalar(sps, int=True, positive=True)  # Samples per symbol

        if isinstance(pulse_shape, str):
            verify_literal(pulse_shape, ["rect", "rc", "srrc"])
            if pulse_shape == "rect":
                if span is None:
                    span = 1
                self._pulse_shape = rectangular(self.sps, span=span)
            elif pulse_shape == "rc":
                if span is None:
                    span = 10
                if alpha is None:
                    alpha = 0.2
                self._pulse_shape = raised_cosine(alpha, span, self.sps)
            elif pulse_shape == "srrc":
                if span is None:
                    span = 10
                if alpha is None:
                    alpha = 0.2
                self._pulse_shape = root_raised_cosine(alpha, span, self.sps)
        else:
            self._pulse_shape = verify_arraylike(pulse_shape, float=True, ndim=1)  # Pulse shape

        self._tx_filter = Interpolator(self.sps, self.pulse_shape)  # Transmit pulse shaping filter
        self._rx_filter = Decimator(self.sps, self.pulse_shape[::-1].conj())  # Receive matched filter

    def __repr__(self) -> str:
        return f"sdr.{type(self).__name__}({self.symbol_map.tolist()}, phase_offset={self.phase_offset})"

    def __str__(self) -> str:
        string = f"sdr.{type(self).__name__}:"
        string += f"\n  order: {self.order}"
        string += f"\n  symbol_map: {self.symbol_map.shape} shape"
        string += f"\n    {self.symbol_map.tolist()}"
        string += f"\n  phase_offset: {self.phase_offset}"
        return string

    def map_symbols(self, s: npt.ArrayLike) -> npt.NDArray[np.complex128]:
        r"""
        Converts the decimal symbols into complex symbols.

        Arguments:
            s: The decimal symbols $s[k]$ to map, $0$ to $M-1$.

        Returns:
            The complex symbols $a[k]$.
        """
        s = verify_arraylike(s, int=True)  # Decimal symbols
        a = self._map_symbols(s)  # Complex symbols
        return convert_output(a)

    def _map_symbols(self, s: npt.NDArray[np.int_]) -> npt.NDArray[np.complex128]:
        a = self.symbol_map[s]  # Complex symbols
        return a

    def decide_symbols(self, a_tilde: npt.ArrayLike) -> tuple[npt.NDArray[np.int_], npt.NDArray[np.complex128]]:
        r"""
        Converts the received complex symbols into MLE symbol decisions.

        This method uses maximum-likelihood estimation (MLE).

        Arguments:
            a_tilde: The received complex symbols $\tilde{a}[k]$.

        Returns:
            - The decimal symbol decisions $\hat{s}[k]$, $0$ to $M-1$.
            - The complex symbol decisions $\hat{a}[k]$.
        """
        a_tilde = verify_arraylike(a_tilde, complex=True)  # Complex symbols
        s_hat, a_hat = self._decide_symbols(a_tilde)  # Decimal and complex symbol decisions
        return convert_output(s_hat), convert_output(a_hat)

    def _decide_symbols(
        self, a_tilde: npt.NDArray[np.complex128]
    ) -> tuple[npt.NDArray[np.int_], npt.NDArray[np.complex128]]:
        error_vectors = np.subtract.outer(a_tilde, self.symbol_map)
        s_hat = np.argmin(np.abs(error_vectors), axis=-1)
        a_hat = self.symbol_map[s_hat]
        return s_hat, a_hat

    def modulate(self, s: npt.ArrayLike) -> npt.NDArray[np.complex128]:
        r"""
        Modulates the decimal symbols into pulse-shaped complex samples.

        Arguments:
            s: The decimal symbols $s[k]$ to modulate, $0$ to $M-1$.

        Returns:
            The pulse-shaped complex samples $x[n]$ with :obj:`sps` samples per symbol
            and length `sps * s.size + pulse_shape.size - 1`.
        """
        s = verify_arraylike(s, int=True)  # Decimal symbols
        return self._modulate(s)

    def _modulate(self, s: npt.NDArray[np.int_]) -> npt.NDArray[np.complex128]:
        a = self._map_symbols(s)  # Complex symbols
        x = self._tx_pulse_shape(a)  # Complex samples
        return x

    def _tx_pulse_shape(self, a: npt.NDArray[np.complex128]) -> npt.NDArray[np.complex128]:
        x = self._tx_filter(a, mode="full")  # Complex samples
        return x

    def demodulate(
        self,
        x_tilde: npt.ArrayLike,
    ) -> tuple[npt.NDArray[np.int_], npt.NDArray[np.complex128], npt.NDArray[np.complex128]]:
        r"""
        Demodulates the pulse-shaped complex samples.

        This method uses matched filtering and maximum-likelihood estimation.

        Arguments:
            x_tilde: The received pulse-shaped complex samples $\tilde{x}[n]$ to demodulate, with :obj:`sps`
                samples per symbol and length `sps * s_hat.size + pulse_shape.size - 1`.

        Returns:
            - The decimal symbol decisions $\hat{s}[k]$, $0$ to $M-1$.
            - The matched filter outputs $\tilde{a}[k]$.
            - The complex symbol decisions $\hat{a}[k]$.
        """
        x_tilde = verify_arraylike(x_tilde, complex=True)  # Complex samples
        # Decimal symbol decisions, complex symbols, complex symbol decisions
        s_hat, a_tilde, a_hat = self._demodulate(x_tilde)
        return convert_output(s_hat), convert_output(a_tilde), convert_output(a_hat)

    def _demodulate(
        self,
        x_tilde: npt.NDArray[np.complex128],
    ) -> tuple[npt.NDArray[np.int_], npt.NDArray[np.complex128], npt.NDArray[np.complex128]]:
        a_tilde = self._rx_matched_filter(x_tilde)  # Complex symbols
        s_hat, a_hat = self._decide_symbols(a_tilde)  # Decimal and complex symbol decisions
        return s_hat, a_tilde, a_hat

    def _rx_matched_filter(
        self,
        x_tilde: npt.NDArray[np.complex128],
    ) -> npt.NDArray[np.complex128]:
        if self.pulse_shape.size % self.sps == 0:
            x_tilde = np.insert(x_tilde, 0, 0)

        a_tilde = self._rx_filter(x_tilde, mode="full")  # Complex symbols

        span = self.pulse_shape.size // self.sps
        if span == 1:
            N_symbols = x_tilde.size // self.sps
            offset = span
        else:
            N_symbols = x_tilde.size // self.sps - span
            offset = span

        # Select the symbol decisions from the output of the decimating filter
        a_tilde = a_tilde[offset : offset + N_symbols]

        return a_tilde

    @abc.abstractmethod
    def ber(self, ebn0: npt.ArrayLike) -> npt.NDArray[np.float64]:
        r"""
        Computes the bit error rate (BER) at the provided $E_b/N_0$ values.

        Arguments:
            ebn0: Bit energy $E_b$ to noise PSD $N_0$ ratio in dB.

        Returns:
            The bit error rate $P_b$.

        See Also:
            sdr.esn0_to_ebn0, sdr.snr_to_ebn0
        """
        raise NotImplementedError("Bit error rate calculation for arbitrary linear modulations is not supported.")

    @abc.abstractmethod
    def ser(self, esn0: npt.ArrayLike) -> npt.NDArray[np.float64]:
        r"""
        Computes the symbol error rate (SER) at the provided $E_s/N_0$ values.

        Arguments:
            esn0: Symbol energy $E_s$ to noise PSD $N_0$ ratio in dB.

        Returns:
            The symbol error rate $P_e$.

        See Also:
            sdr.ebn0_to_esn0, sdr.snr_to_esn0
        """
        raise NotImplementedError("Symbol error rate calculation for arbitrary linear modulations is not supported.")

    @property
    def order(self) -> int:
        r"""
        The modulation order $M = 2^k$.
        """
        return self._order

    @property
    def bps(self) -> int:
        r"""
        The number of coded bits per symbol $k = \log_2 M$.
        """
        return self._bps

    @property
    def phase_offset(self) -> float:
        r"""
        The phase offset $\phi$ in degrees.
        """
        return self._phase_offset

    @property
    def symbol_map(self) -> npt.NDArray[np.complex128]:
        r"""
        The symbol map $\{0, \dots, M-1\} \mapsto \mathbb{C}$.

        This maps decimal symbols from $0$ to $M-1$ to complex symbols.
        """
        return self._symbol_map

    @property
    def sps(self) -> int:
        r"""
        The number of samples per symbol $f_s / f_{sym}$.
        """
        return self._sps

    @property
    def pulse_shape(self) -> npt.NDArray[np.float64]:
        r"""
        The pulse shape $h[n]$ of the modulated signal.
        """
        return self._pulse_shape

    @property
    def tx_filter(self) -> Interpolator:
        r"""
        The transmit interpolating pulse shaping filter.

        The filter coefficients are the pulse shape $h[n]$.
        """
        return self._tx_filter

    @property
    def rx_filter(self) -> Decimator:
        r"""
        The receive decimating matched filter.

        The filter coefficients are matched to the pulse shape $h[-n]^*$.
        """
        return self._rx_filter
